from __future__ import annotations
from collections.abc import Sequence

import numpy

from ptychodus.api.object import Object, ObjectGeometryProvider
from ptychodus.api.object_gen import generate_random_object

from .builder import ObjectBuilder
from .settings import ObjectSettings


class RandomObjectBuilder(ObjectBuilder):
    def __init__(self, rng: numpy.random.Generator, settings: ObjectSettings) -> None:
        super().__init__(settings, 'random')
        self._rng = rng
        self._settings = settings

        self.amplitude_mean = settings.amplitude_mean.copy()
        self._add_parameter('amplitude_mean', self.amplitude_mean)
        self.amplitude_deviation = settings.amplitude_deviation.copy()
        self._add_parameter('amplitude_deviation', self.amplitude_deviation)
        self.phase_deviation_tr = settings.phase_deviation_tr.copy()
        self._add_parameter('phase_deviation_turns', self.phase_deviation_tr)
        self.blur_deviation_px = settings.blur_deviation_px.copy()
        self._add_parameter('blur_deviation_px', self.blur_deviation_px)

    def copy(self) -> RandomObjectBuilder:
        builder = RandomObjectBuilder(self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        object_ = generate_random_object(
            self._rng,
            geometry_provider.get_object_geometry(),
            amplitude_mean=self.amplitude_mean.get_value(),
            amplitude_deviation=self.amplitude_deviation.get_value(),
            phase_mean=0.0,
            phase_deviation_tr=self.phase_deviation_tr.get_value(),
            blur_deviation_px=self.blur_deviation_px.get_value(),
        )
        return self._create_object(object_, layer_spacing_m)



from __future__ import annotations

import numpy

from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider
from ptychodus.api.probe_gen import generate_average_pattern_probe, rescale_probe_intensity

from ...diffraction import DiffractionAPI
from .builder import ProbeSequenceBuilder
from .settings import ProbeSettings


class AveragePatternProbeBuilder(ProbeSequenceBuilder):
    def __init__(
        self,
        rng: numpy.random.Generator,
        settings: ProbeSettings,
        diffraction_api: DiffractionAPI,
    ) -> None:
        super().__init__(settings, 'average_pattern')
        self._rng = rng
        self._settings = settings
        self._diffraction_api = diffraction_api

    def copy(self) -> AveragePatternProbeBuilder:
        builder = AveragePatternProbeBuilder(self._rng, self._settings, self._diffraction_api)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        probe = rescale_probe_intensity(
            generate_average_pattern_probe(
                geometry_provider.get_probe_geometry(),
                self._diffraction_api.get_assembled_data(),
                probe_wavelength_m=geometry_provider.probe_wavelength_m,
                detector_distance_m=geometry_provider.detector_distance_m,
            ),
            geometry_provider.probe_photon_count,
        )
        return self._build_probe_modes(self._rng, probe, geometry_provider.num_scan_points)



from __future__ import annotations
import logging

from ptychodus.api.observer import Observable
from ptychodus.api.parametric import ParameterGroup
from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider

from .builder import FromMemoryProbeBuilder, ProbeSequenceBuilder
from .settings import ProbeSettings

logger = logging.getLogger(__name__)


class ProbeRepositoryItem(ParameterGroup):
    def __init__(
        self,
        geometry_provider: ProbeGeometryProvider,
        settings: ProbeSettings,
        builder: ProbeSequenceBuilder,
    ) -> None:
        super().__init__()
        self._geometry_provider = geometry_provider
        self._settings = settings
        self._builder = builder
        self._probe_seq = ProbeSequence(array=None, opr_weights=None, pixel_geometry=None)

        self._add_group('builder', builder, observe=True)
        self._rebuild()

    def assign_item(self, item: ProbeRepositoryItem) -> None:
        self.set_builder(item.get_builder().copy())
        self._rebuild()

    def assign(self, probe: ProbeSequence) -> None:
        builder = FromMemoryProbeBuilder(self._settings, probe)
        self.set_builder(builder)

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

        self._builder.sync_to_settings()

    def get_probes(self) -> ProbeSequence:
        return self._probe_seq

    def get_builder(self) -> ProbeSequenceBuilder:
        return self._builder

    def set_builder(self, builder: ProbeSequenceBuilder) -> None:
        group = 'builder'
        self._remove_group(group)
        self._builder.remove_observer(self)
        self._builder = builder
        self._builder.add_observer(self)
        self._add_group(group, self._builder, observe=True)
        self._rebuild()

    def _rebuild(self) -> None:
        try:
            probe_seq = self._builder.build(self._geometry_provider)
        except Exception:
            logger.exception('Failed to rebuild probe!')
        else:
            self._probe_seq = probe_seq
            self.notify_observers()

    def _update(self, observable: Observable) -> None:
        if observable is self._builder:
            self._rebuild()
        else:
            super()._update(observable)



from collections.abc import Iterable, Sequence
from dataclasses import dataclass
import logging

import numpy
import scipy.linalg

from .common import ComplexArrayType, RealArrayType
from .geometry import PixelGeometry, ZernikeMonomial
from .probe import Probe, ProbeGeometry, ProbeSequence
from .propagator import (
    AngularSpectrumPropagator,
    FresnelTransformPropagator,
    PropagatorParameters,
    intensity,
)
from .reconstructor import AssembledDiffractionData


logger = logging.getLogger(__name__)


def rescale_probe_intensity(probe: Probe, new_intensity: float) -> Probe:
    array = probe.get_array()
    old_intensity = numpy.sum(intensity(array))

    if new_intensity <= 0:
        logger.warning('Refusing to rescale probe to zero intensity!')
    elif numpy.isnan(old_intensity):
        logger.warning('Cannot rescale probe with NaN values!')
    elif old_intensity <= 0:
        logger.warning('Cannot rescale probe with zero intensity!')
    else:
        return Probe(
            array=array * numpy.sqrt(new_intensity / old_intensity),
            pixel_geometry=probe.get_pixel_geometry(),
        )

    return probe


def defocus_probe(
    probe: Probe,
    *,
    probe_wavelength_m: float,
    defocus_distance_m: float,
) -> Probe:
    pixel_geometry = probe.get_pixel_geometry()
    propagator_parameters = PropagatorParameters(
        wavelength_m=probe_wavelength_m,
        width_px=probe.width_px,
        height_px=probe.height_px,
        pixel_width_m=pixel_geometry.width_m,
        pixel_height_m=pixel_geometry.height_m,
        propagation_distance_m=defocus_distance_m,
    )
    propagator = AngularSpectrumPropagator(propagator_parameters)
    return Probe(
        array=propagator.propagate(probe.get_array()),
        pixel_geometry=pixel_geometry,
    )


def generate_disk_probe(
    geometry: ProbeGeometry,
    *,
    radius_m: float,
) -> Probe:
    coords = geometry.get_transverse_coordinates()
    return Probe(
        array=numpy.where(coords.position_r_m < radius_m, 1, 0) + 0j,
        pixel_geometry=geometry.get_pixel_geometry(),
    )


def generate_rectangular_probe(
    geometry: ProbeGeometry,
    *,
    width_m: float,
    height_m: float,
) -> Probe:
    coords = geometry.get_transverse_coordinates()
    is_inside = numpy.logical_and(
        numpy.fabs(coords.position_x_m) < 0.5 * width_m,
        numpy.fabs(coords.position_y_m) < 0.5 * height_m,
    )
    return Probe(
        array=numpy.where(is_inside, 1, 0) + 0j,
        pixel_geometry=geometry.get_pixel_geometry(),
    )


def generate_super_gaussian_probe(
    geometry: ProbeGeometry,
    *,
    annular_radius_m: float,
    fwhm_m: float,
    order_parameter: float,
) -> Probe:
    coords = geometry.get_transverse_coordinates()
    z = (coords.position_r_m - annular_radius_m) / fwhm_m
    zp = numpy.power(2 * z, 2 * order_parameter)
    return Probe(
        array=numpy.exp(-numpy.log(2) * zp) + 0j,
        pixel_geometry=geometry.get_pixel_geometry(),
    )


def generate_average_pattern_probe(
    geometry: ProbeGeometry,
    assembled_data: AssembledDiffractionData,
    *,
    probe_wavelength_m: float,
    detector_distance_m: float,
) -> Probe:
    detector_intensity = numpy.mean(assembled_data.get_patterns(), axis=0)
    detector_pixel_geometry = assembled_data.get_pixel_geometry()
    propagator_parameters = PropagatorParameters(
        wavelength_m=probe_wavelength_m,
        width_px=detector_intensity.shape[-1],
        height_px=detector_intensity.shape[-2],
        pixel_width_m=detector_pixel_geometry.width_m,
        pixel_height_m=detector_pixel_geometry.height_m,
        propagation_distance_m=-detector_distance_m,
    )
    propagator = FresnelTransformPropagator(propagator_parameters)
    array = propagator.propagate(numpy.sqrt(detector_intensity).astype(complex))

    return Probe(
        array=array,
        pixel_geometry=geometry.get_pixel_geometry(),
    )


@dataclass(frozen=True)
class FresnelZonePlate:
    """Physical parameters of a Fresnel zone plate optic."""

    zone_plate_diameter_m: float
    outermost_zone_width_m: float
    central_beamstop_diameter_m: float

    def get_focal_length_m(self, central_wavelength_m: float) -> float:
        return self.zone_plate_diameter_m * self.outermost_zone_width_m / central_wavelength_m


def generate_fresnel_zone_plate_probe(
    geometry: ProbeGeometry,
    zone_plate: FresnelZonePlate,
    *,
    probe_wavelength_m: float,
    defocus_distance_m: float,
) -> Probe:
    focal_length_m = zone_plate.get_focal_length_m(probe_wavelength_m)
    propagation_distance_m = focal_length_m + defocus_distance_m

    fzp_half_width = (geometry.width_px + 1) // 2
    fzp_half_height = (geometry.height_px + 1) // 2
    fzp_plane_pixel_size_numerator = probe_wavelength_m * propagation_distance_m
    fzp_pixel_geometry = PixelGeometry(
        width_m=fzp_plane_pixel_size_numerator / geometry.width_m,
        height_m=fzp_plane_pixel_size_numerator / geometry.height_m,
    )

    # coordinate on FZP plane
    lx_fzp = -fzp_pixel_geometry.width_m * numpy.arange(-fzp_half_width, fzp_half_width)
    ly_fzp = -fzp_pixel_geometry.height_m * numpy.arange(-fzp_half_height, fzp_half_height)

    XX_FZP, YY_FZP = numpy.meshgrid(lx_fzp, ly_fzp)  # noqa: N806
    RR_FZP = numpy.hypot(XX_FZP, YY_FZP)  # noqa: N806

    # transmission function of FZP
    T = numpy.exp(  # noqa: N806
        -2j * numpy.pi / probe_wavelength_m * (XX_FZP**2 + YY_FZP**2) / 2 / focal_length_m
    )
    C = RR_FZP <= zone_plate.zone_plate_diameter_m / 2  # noqa: N806
    H = RR_FZP >= zone_plate.central_beamstop_diameter_m / 2  # noqa: N806
    fzp_transmission_function = T * C * H

    propagator_parameters = PropagatorParameters(
        wavelength_m=probe_wavelength_m,
        width_px=fzp_transmission_function.shape[-1],
        height_px=fzp_transmission_function.shape[-2],
        pixel_width_m=fzp_pixel_geometry.width_m,
        pixel_height_m=fzp_pixel_geometry.height_m,
        propagation_distance_m=propagation_distance_m,
    )
    propagator = FresnelTransformPropagator(propagator_parameters)

    return Probe(
        array=propagator.propagate(fzp_transmission_function),
        pixel_geometry=geometry.get_pixel_geometry(),
    )


def generate_zernike_probe(
    geometry: ProbeGeometry, polynomial: Iterable[ZernikeMonomial], *, radius_m: float
) -> Probe:
    coords = geometry.get_transverse_coordinates()
    distance = coords.position_r_m / radius_m
    angle_rad = coords.angle_rad
    array = numpy.zeros_like(distance, dtype=complex)

    for monomial in polynomial:
        array += monomial(distance, angle_rad)

    return Probe(
        array=array,
        pixel_geometry=geometry.get_pixel_geometry(),
    )


def _random_phase_shift_axis(rng: numpy.random.Generator, size: int) -> ComplexArrayType:
    a = rng.uniform() - 0.5
    b = (size - 1 - 2 * numpy.arange(size)) / size
    return numpy.exp(1j * numpy.pi * a * b)


def generate_incoherent_probe_modes(
    rng: numpy.random.Generator,
    probe: Probe,
    imode_weights: Sequence[float],
    *,
    orthogonalize: bool = True,
) -> Probe:
    num_imodes = len(imode_weights)
    array_in = probe.get_array()

    if numpy.isnan(array_in).any():
        logger.warning('Probe without incoherent modes contains NaN values!')
        return probe

    array_out_shape = num_imodes, *array_in.shape[-2:]
    array_out = numpy.zeros(array_out_shape, dtype=array_in.dtype)

    for imode in range(num_imodes):
        if imode < array_in.shape[-3]:
            # preserve existing incoherent modes
            values = array_in[imode, :, :]
        else:
            # apply random phase shift to 1st incoherent mode
            phase_shift_y = _random_phase_shift_axis(rng, array_in.shape[-2])
            phase_shift_x = _random_phase_shift_axis(rng, array_in.shape[-1])
            values = array_in[0, :, :] * numpy.outer(phase_shift_y, phase_shift_x)

        array_out[imode, :, :] = values

    if orthogonalize and array_in.shape[-3] > 1:
        imodes_as_rows = array_out.reshape(num_imodes, -1)
        imodes_as_ortho_rows = scipy.linalg.orth(imodes_as_rows.T).T
        array_out = imodes_as_ortho_rows.reshape(array_out_shape)

    if numpy.isnan(array_out).any():
        logger.warning('Probe with incoherent modes contains NaN values!')
        return probe

    normalized_imode_weights = numpy.asarray(imode_weights) / numpy.sum(imode_weights)
    imode_intensity = numpy.sum(intensity(array_in)) * normalized_imode_weights

    for imode, intensity_out in enumerate(imode_intensity):
        intensity_in = numpy.sum(intensity(array_out[imode, :, :]))

        if intensity_in <= 0:
            logger.warning('Cannot rescale imode with zero intensity!')
        else:
            array_out[imode, :, :] *= numpy.sqrt(intensity_out / intensity_in)

    return Probe(
        array=array_out,
        pixel_geometry=probe.get_pixel_geometry(),
    )


def generate_coherent_probe_modes(
    rng: numpy.random.Generator,
    probe: Probe,
    *,
    num_cmodes: int,
    num_diffraction_patterns: int,
    small_value: float = 1.0e-6,
    normalize_cmodes: bool = True,
) -> ProbeSequence:
    opr_weights: RealArrayType | None = None

    if num_cmodes > 1:
        opr_weights = small_value * rng.normal(size=(num_diffraction_patterns, num_cmodes))
        opr_weights[:, 0] = 1.0

    array_in = probe.get_array()
    array_out = numpy.zeros((num_cmodes, *array_in.shape), array_in.dtype)
    array_out[0, :, :, :] = array_in[:, :, :]

    for cmode in range(1, num_cmodes):
        real = rng.normal(size=array_out.shape[-2:])
        imag = rng.normal(size=array_out.shape[-2:])
        values = real + 1j * imag

        if normalize_cmodes:
            values /= numpy.sqrt(numpy.mean(intensity(values)))

        array_out[cmode, 0, :, :] = values

    return ProbeSequence(
        array=array_out,
        opr_weights=opr_weights,
        pixel_geometry=probe.get_pixel_geometry(),
    )



from bisect import bisect
from collections.abc import Sequence
from datetime import datetime, timezone
from typing import overload
import queue

from ptychodus.api.observer import ObservableSequence

from .settings import GlobusSettings
from .client import GlobusClient, GlobusStatus


class GlobusStatusRepository(ObservableSequence[GlobusStatus]):
    def __init__(
        self, settings: GlobusSettings, client: GlobusClient, status_q: queue.Queue[GlobusStatus]
    ) -> None:
        super().__init__()
        self._settings = settings
        self._client = client
        self._status_q = status_q
        self._status_list: list[GlobusStatus] = list()
        self._run_id_to_index_map: dict[str, int] = dict()
        self._status_date_time = datetime.min

    def refresh_status(self) -> None:
        self._status_date_time = datetime.now(timezone.utc)
        self._client.refresh_status()

    @overload
    def __getitem__(self, index: int) -> GlobusStatus: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[GlobusStatus]: ...

    def __getitem__(self, index: int | slice) -> GlobusStatus | Sequence[GlobusStatus]:
        return self._status_list[index]

    def __len__(self) -> int:
        return len(self._status_list)

    def _update_run_id_to_index_map(self) -> None:
        self._run_id_to_index_map = {
            status.run_id: index for index, status in enumerate(self._status_list)
        }

    def run_foreground_tasks(self) -> None:
        if self._settings.status_auto_refresh.get_value():
            status_time_delta = datetime.now(timezone.utc) - self._status_date_time
            status_age_s = status_time_delta.total_seconds()

            if status_age_s >= self._settings.status_refresh_interval_s.get_value():
                self.refresh_status()

        while True:
            try:
                new_status = self._status_q.get(block=False)
            except queue.Empty:
                break
            else:
                try:
                    index = self._run_id_to_index_map[new_status.run_id]
                except KeyError:
                    index = bisect(
                        self._status_list, new_status.start_time, key=lambda x: x.start_time
                    )
                    self._status_list.insert(index, new_status)
                    self._update_run_id_to_index_map()
                    self.notify_observers_item_inserted(index, new_status)
                else:
                    old_status = self._status_list[index]

                    if old_status != new_status:
                        self._status_list[index] = new_status
                        self.notify_observers_item_changed(index, new_status)

                self._status_q.task_done()



from __future__ import annotations
from collections.abc import Iterator

import numpy

from ptychodus.api.plugins import PluginChooser
from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider
from ptychodus.api.probe_gen import (
    FresnelZonePlate,
    generate_fresnel_zone_plate_probe,
    rescale_probe_intensity,
)

from .builder import ProbeSequenceBuilder
from .settings import ProbeSettings


class FresnelZonePlateProbeBuilder(ProbeSequenceBuilder):
    def __init__(
        self,
        rng: numpy.random.Generator,
        settings: ProbeSettings,
        fresnel_zone_plate_chooser: PluginChooser[FresnelZonePlate],
    ) -> None:
        super().__init__(settings, 'fresnel_zone_plate')
        self._rng = rng
        self._settings = settings
        self._fresnel_zone_plate_chooser = fresnel_zone_plate_chooser

        self.zone_plate_diameter_m = settings.zone_plate_diameter_m.copy()
        self._add_parameter('zone_plate_diameter_m', self.zone_plate_diameter_m)

        self.outermost_zone_width_m = settings.outermost_zone_width_m.copy()
        self._add_parameter('outermost_zone_width_m', self.outermost_zone_width_m)

        self.central_beamstop_diameter_m = settings.central_beamstop_diameter_m.copy()
        self._add_parameter('central_beamstop_diameter_m', self.central_beamstop_diameter_m)

        # from sample to the focal plane
        self.defocus_distance_m = settings.defocus_distance_m.copy()
        self._add_parameter('defocus_distance_m', self.defocus_distance_m)

    def copy(self) -> FresnelZonePlateProbeBuilder:
        builder = FresnelZonePlateProbeBuilder(
            self._rng, self._settings, self._fresnel_zone_plate_chooser
        )

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def labels_for_presets(self) -> Iterator[str]:
        for plugin in self._fresnel_zone_plate_chooser:
            yield plugin.display_name

    def apply_presets(self, display_name: str) -> None:
        self._fresnel_zone_plate_chooser.set_current_plugin(display_name)
        fzp = self._fresnel_zone_plate_chooser.get_current_plugin().strategy
        self.zone_plate_diameter_m.set_value(fzp.zone_plate_diameter_m)
        self.outermost_zone_width_m.set_value(fzp.outermost_zone_width_m)
        self.central_beamstop_diameter_m.set_value(fzp.central_beamstop_diameter_m)

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        zone_plate = FresnelZonePlate(
            zone_plate_diameter_m=self.zone_plate_diameter_m.get_value(),
            outermost_zone_width_m=self.outermost_zone_width_m.get_value(),
            central_beamstop_diameter_m=self.central_beamstop_diameter_m.get_value(),
        )
        probe = rescale_probe_intensity(
            generate_fresnel_zone_plate_probe(
                geometry=geometry_provider.get_probe_geometry(),
                zone_plate=zone_plate,
                probe_wavelength_m=geometry_provider.probe_wavelength_m,
                defocus_distance_m=self.defocus_distance_m.get_value(),
            ),
            geometry_provider.probe_photon_count,
        )
        return self._build_probe_modes(self._rng, probe, geometry_provider.num_scan_points)



from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from datetime import datetime
from typing import Any


@dataclass(frozen=True)
class GlobusJob:
    flow_input: Mapping[str, Any]
    label: str
    tags: Sequence[str]


@dataclass(frozen=True)
class GlobusStatus:
    label: str
    start_time: datetime
    completion_time: datetime | None
    status: str
    action: str
    run_id: str

    @property
    def run_url(self) -> str:
        return f'https://app.globus.org/runs/{self.run_id}/logs'


class GlobusClient(ABC):
    @property
    @abstractmethod
    def is_supported(self) -> bool:
        pass

    @abstractmethod
    def start(self) -> None:
        pass

    @abstractmethod
    def stop(self) -> None:
        pass

    @abstractmethod
    def run_flow(self, job: GlobusJob) -> None:
        pass

    @abstractmethod
    def refresh_status(self) -> None:
        pass


class FakeGlobusClient(GlobusClient):
    @property
    def is_supported(self) -> bool:
        return False

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass

    def run_flow(self, job: GlobusJob) -> None:
        pass

    def refresh_status(self) -> None:
        pass


class RunGlobusFlow:
    def __init__(self, client: GlobusClient, job: GlobusJob) -> None:
        self._client = client
        self._job = job

    def __call__(self) -> None:
        self._client.run_flow(self._job)



from __future__ import annotations

import numpy

from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider
from ptychodus.api.probe_gen import defocus_probe, generate_disk_probe, rescale_probe_intensity

from .builder import ProbeSequenceBuilder
from .settings import ProbeSettings


class DiskProbeBuilder(ProbeSequenceBuilder):
    def __init__(self, rng: numpy.random.Generator, settings: ProbeSettings) -> None:
        super().__init__(settings, 'disk')
        self._rng = rng
        self._settings = settings

        self.diameter_m = settings.disk_diameter_m.copy()
        self._add_parameter('diameter_m', self.diameter_m)

        # from sample to the focal plane
        self.defocus_distance_m = settings.defocus_distance_m.copy()
        self._add_parameter('defocus_distance_m', self.defocus_distance_m)

    def copy(self) -> DiskProbeBuilder:
        builder = DiskProbeBuilder(self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        probe = rescale_probe_intensity(
            defocus_probe(
                generate_disk_probe(
                    geometry_provider.get_probe_geometry(),
                    radius_m=self.diameter_m.get_value() / 2.0,
                ),
                probe_wavelength_m=geometry_provider.probe_wavelength_m,
                defocus_distance_m=self.defocus_distance_m.get_value(),
            ),
            geometry_provider.probe_photon_count,
        )
        return self._build_probe_modes(self._rng, probe, geometry_provider.num_scan_points)



from __future__ import annotations
from enum import IntEnum

import numpy

from ptychodus.api.probe_positions import ProbePositionSequence
from ptychodus.api.probe_positions_gen import generate_cartesian_probe_positions

from .builder import ProbePositionsBuilder
from .settings import ProbePositionsSettings


class CartesianProbePositionsVariant(IntEnum):
    RECTANGULAR_RASTER = 0x0
    RECTANGULAR_SNAKE = 0x1
    TRIANGULAR_RASTER = 0x2
    TRIANGULAR_SNAKE = 0x3
    SQUARE_RASTER = 0x4
    SQUARE_SNAKE = 0x5
    HEXAGONAL_RASTER = 0x6
    HEXAGONAL_SNAKE = 0x7

    @property
    def is_snaked(self) -> bool:
        return self.value & 1 != 0

    @property
    def is_staggered(self) -> bool:
        return self.value & 2 != 0

    @property
    def is_equilateral(self) -> bool:
        return self.value & 4 != 0


class CartesianProbePositionsBuilder(ProbePositionsBuilder):
    def __init__(
        self,
        variant: CartesianProbePositionsVariant,
        rng: numpy.random.Generator,
        settings: ProbePositionsSettings,
    ) -> None:
        super().__init__(rng, settings, variant.name.lower())
        self._rng = rng
        self._variant = variant
        self._settings = settings

        self.num_points_x = settings.num_points_x.copy()
        self._add_parameter('num_points_x', self.num_points_x)

        self.num_points_y = settings.num_points_y.copy()
        self._add_parameter('num_points_y', self.num_points_y)

        self.step_size_x_m = settings.step_size_x_m.copy()
        self._add_parameter('step_size_x_m', self.step_size_x_m)

        self.step_size_y_m = settings.step_size_y_m.copy()
        self._add_parameter('step_size_y_m', self.step_size_y_m)

    def copy(self) -> CartesianProbePositionsBuilder:
        builder = CartesianProbePositionsBuilder(self._variant, self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    @property
    def is_equilateral(self) -> bool:
        return self._variant.is_equilateral

    def build(self) -> ProbePositionSequence:
        step_size_x_m = self.step_size_x_m.get_value()

        if self._variant.is_equilateral:
            step_size_y_m = step_size_x_m

            if self._variant.is_staggered:
                step_size_y_m *= numpy.sqrt(0.75)
        else:
            step_size_y_m = self.step_size_y_m.get_value()

        positions = generate_cartesian_probe_positions(
            self.num_points_x.get_value(),
            self.num_points_y.get_value(),
            step_size_x_m,
            step_size_y_m,
            snake=self._variant.is_snaked,
            stagger=self._variant.is_staggered,
        )
        return self._create_position_sequence(positions)



from __future__ import annotations

import numpy

from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider
from ptychodus.api.probe_gen import (
    defocus_probe,
    generate_rectangular_probe,
    rescale_probe_intensity,
)

from .builder import ProbeSequenceBuilder
from .settings import ProbeSettings


class RectangularProbeBuilder(ProbeSequenceBuilder):
    def __init__(self, rng: numpy.random.Generator, settings: ProbeSettings) -> None:
        super().__init__(settings, 'rectangular')
        self._rng = rng
        self._settings = settings

        self.width_m = settings.rectangle_width_m.copy()
        self._add_parameter('width_m', self.width_m)

        self.height_m = settings.rectangle_height_m.copy()
        self._add_parameter('height_m', self.height_m)

        # from sample to the focal plane
        self.defocus_distance_m = settings.defocus_distance_m.copy()
        self._add_parameter('defocus_distance_m', self.defocus_distance_m)

    def copy(self) -> RectangularProbeBuilder:
        builder = RectangularProbeBuilder(self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        probe = rescale_probe_intensity(
            defocus_probe(
                generate_rectangular_probe(
                    geometry_provider.get_probe_geometry(),
                    width_m=self.width_m.get_value(),
                    height_m=self.height_m.get_value(),
                ),
                probe_wavelength_m=geometry_provider.probe_wavelength_m,
                defocus_distance_m=self.defocus_distance_m.get_value(),
            ),
            geometry_provider.probe_photon_count,
        )
        return self._build_probe_modes(self._rng, probe, geometry_provider.num_scan_points)



from __future__ import annotations
from collections.abc import Sequence

import numpy

from ptychodus.api.object import Object, ObjectGeometryProvider
from ptychodus.api.object_gen import generate_gaussian_random_field_object

from .builder import ObjectBuilder
from .settings import ObjectSettings


class GaussianRandomFieldObjectBuilder(ObjectBuilder):
    def __init__(self, rng: numpy.random.Generator, settings: ObjectSettings) -> None:
        super().__init__(settings, 'gaussian_random_field')
        self._rng = rng
        self._settings = settings

        self.correlation_length_px = settings.correlation_length_px.copy()
        self._add_parameter('correlation_length_px', self.correlation_length_px)

    def copy(self) -> GaussianRandomFieldObjectBuilder:
        builder = GaussianRandomFieldObjectBuilder(self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        object_ = generate_gaussian_random_field_object(
            self._rng,
            geometry_provider.get_object_geometry(),
            correlation_length_px=self.correlation_length_px.get_value(),
        )
        return self._create_object(object_, layer_spacing_m)



from __future__ import annotations
from collections.abc import Sequence
import logging


from ptychodus.api.object import Object, ObjectGeometryProvider
from ptychodus.api.object_gen import generate_stxm_object

from ...diffraction import DiffractionAPI
from .builder import ObjectBuilder
from .settings import ObjectSettings

logger = logging.getLogger(__name__)


class STXMObjectBuilder(ObjectBuilder):
    def __init__(
        self,
        settings: ObjectSettings,
        diffraction_api: DiffractionAPI,
    ) -> None:
        super().__init__(settings, 'stxm')
        self._settings = settings
        self._diffraction_api = diffraction_api

    def copy(self) -> STXMObjectBuilder:
        builder = STXMObjectBuilder(self._settings, self._diffraction_api)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        object_ = generate_stxm_object(
            geometry_provider.get_object_geometry(),
            self._diffraction_api.get_assembled_data(),
            geometry_provider.get_probe_positions(),
        )
        return self._create_object(object_, layer_spacing_m)



from __future__ import annotations
from datetime import datetime
from pprint import pformat
import json
import logging
import queue
import threading

import fair_research_login


from ..task_manager import TaskManager
from ._gladier_client import create_client as create_ptychodus_client
from .authorizer import GlobusAuthorizer
from .client import GlobusClient, GlobusJob, GlobusStatus
from .settings import GlobusSettings

__all__ = ['RealGlobusClient']

logger = logging.getLogger(__name__)


class CustomCodeHandler(fair_research_login.CodeHandler):
    def __init__(
        self,
        auth_url_q: queue.Queue[str],
        auth_code_q: queue.Queue[str],
        stop_event: threading.Event,
    ) -> None:
        super().__init__()
        self._auth_url_q = auth_url_q
        self._auth_code_q = auth_code_q
        self._stop_event = stop_event
        self._code = ''
        self.set_browser_enabled(False)

    def authenticate(self, url: str) -> str:
        self._auth_url_q.put(url)

        while not self._stop_event.is_set():
            try:
                self._code = self._auth_code_q.get(block=True, timeout=1.0)
            except queue.Empty:
                pass

        return self._code

    def get_code(self) -> str:
        return self._code


class GlobusClientThread(threading.Thread):
    def __init__(
        self,
        auth_url_q: queue.Queue[str],
        auth_code_q: queue.Queue[str],
        job_q: queue.Queue[GlobusJob],
        status_q: queue.Queue[GlobusStatus],
        refresh_status_event: threading.Event,
        stop_event: threading.Event,
    ) -> None:
        super().__init__()
        self._auth_url_q = auth_url_q
        self._auth_code_q = auth_code_q
        self._job_q = job_q
        self._status_q = status_q
        self._refresh_status_event = refresh_status_event
        self._stop_event = stop_event
        self._gladier_client = create_ptychodus_client(
            [CustomCodeHandler(auth_url_q, auth_code_q, stop_event)]
        )

    def _get_current_action(self, run_id: str) -> str:
        status = self._gladier_client.get_status(run_id)
        action = status.get('state_name')

        if not action:
            try:
                det = status['details']
            except Exception:
                logger.exception('Unexpected flow status!')
                logger.error(pformat(status))
            else:
                if det.get('details') and det['details'].get('state_name'):
                    action = det['details']['state_name']
                elif det.get('details') and det['details'].get('output'):
                    action = list(det['details']['output'].keys())[0]
                elif det.get('action_statuses'):
                    action = det['action_statuses'][0].get('state_name')
                elif det.get('code') == 'FlowStarting':
                    pass

        return action

    def _refresh_status(self) -> None:
        logger.debug('Refreshing status.')

        flows_manager = self._gladier_client.flows_manager
        flow_id = flows_manager.get_flow_id()
        flows_client = flows_manager.flows_client
        response = flows_client.list_runs(filter_flow_id=flow_id)
        run_dict_list = response['runs']

        while response['has_next_page']:
            response = flows_client.list_runs(filter_flow_id=flow_id, marker=response['marker'])
            run_dict_list.extend(response['runs'])

        for run_dict in run_dict_list:
            run_id = run_dict.get('run_id', '')
            action = self._get_current_action(run_id)
            start_time_str = run_dict.get('start_time', '')

            try:
                start_time = datetime.fromisoformat(start_time_str)
            except ValueError:
                logger.warning(f'Failed to parse startTime "{start_time_str}"!')
                start_time = datetime.min

            completion_time_str = run_dict.get('completion_time', '')

            try:
                completion_time = datetime.fromisoformat(completion_time_str)
            except ValueError:
                completion_time = None

            status = GlobusStatus(
                label=run_dict.get('label', ''),
                start_time=start_time,
                completion_time=completion_time,
                status=run_dict.get('status', ''),
                action=action,
                run_id=run_id,
            )
            self._status_q.put(status)

    def run(self) -> None:
        while not self._stop_event.is_set():
            if self._refresh_status_event.is_set():
                self._refresh_status_event.clear()
                self._refresh_status()

            try:
                job = self._job_q.get(block=True, timeout=1)
            except queue.Empty:
                pass
            else:
                try:
                    response = self._gladier_client.run_flow(
                        flow_input={'input': job.flow_input},
                        label=job.label,
                        tags=job.tags,
                    )
                except Exception:
                    logger.exception('Error running flow!')
                else:
                    logger.info(f'Run Flow Response: {json.dumps(response, indent=4)}')
                finally:
                    self._job_q.task_done()


class RealGlobusClient(GlobusClient):
    def __init__(
        self,
        task_manager: TaskManager,
        settings: GlobusSettings,
        authorizer: GlobusAuthorizer,
        status_q: queue.Queue[GlobusStatus],
    ) -> None:
        super().__init__()
        self._task_manager = task_manager
        self._settings = settings
        self._authorizer = authorizer
        self._status_q = status_q

        self._stop_event = threading.Event()
        self._refresh_status_event = threading.Event()
        self._job_q: queue.Queue[GlobusJob] = queue.Queue()
        self._worker: GlobusClientThread | None = None

    @property
    def is_supported(self) -> bool:
        return True

    def start(self) -> None:  # FIXME start as needed
        if self._worker is None:
            logger.debug('Starting Globus thread...')
            self._stop_event.clear()
            self._worker = GlobusClientThread(
                auth_url_q=self._authorizer._auth_url_q,
                auth_code_q=self._authorizer._auth_code_q,
                job_q=self._job_q,
                status_q=self._status_q,
                refresh_status_event=self._refresh_status_event,
                stop_event=self._stop_event,
            )
            self._worker.start()
            logger.debug('Globus thread started.')
        else:
            logger.warning('Worker already started!')

    def stop(self) -> None:
        if self._stop_event.is_set():
            logger.debug('Globus thread already stopped.')
        else:
            logger.debug('Finishing tasks...')
            self._job_q.join()
            logger.debug('Tasks finished.')

            if self._worker is None:
                logger.warning('Worker is None!')
            else:
                logger.debug('Stopping Globus thread...')
                self._stop_event.set()
                self._worker.join()
                self._worker = None
                logger.debug('Globus thread stopped.')

    def run_flow(self, job: GlobusJob) -> None:
        self._job_q.put(job)

    def refresh_status(self) -> None:
        self._refresh_status_event.set()



from __future__ import annotations
from abc import abstractmethod
from collections.abc import Sequence
import logging


from ptychodus.api.object import Object, ObjectFileReader, ObjectGeometryProvider
from ptychodus.api.object_gen import generate_layers, pad_object
from ptychodus.api.parametric import ParameterGroup

from .settings import ObjectSettings

logger = logging.getLogger(__name__)


class ObjectBuilder(ParameterGroup):
    def __init__(self, settings: ObjectSettings, name: str) -> None:
        super().__init__()
        self._name = settings.builder.copy()
        self._name.set_value(name)
        self._add_parameter('name', self._name)

        self.extra_padding_x = settings.extra_padding_x.copy()
        self._add_parameter('extra_padding_x', self.extra_padding_x)
        self.extra_padding_y = settings.extra_padding_y.copy()
        self._add_parameter('extra_padding_y', self.extra_padding_y)

    def get_name(self) -> str:
        return self._name.get_value()

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

    @abstractmethod
    def copy(self) -> ObjectBuilder:
        pass

    @abstractmethod
    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        pass

    def _create_object(
        self,
        object_: Object,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        return pad_object(
            generate_layers(object_, layer_spacing_m),
            self.extra_padding_x.get_value(),
            self.extra_padding_y.get_value(),
        )


class FromMemoryObjectBuilder(ObjectBuilder):
    def __init__(self, settings: ObjectSettings, object_: Object) -> None:
        super().__init__(settings, 'from_memory')
        self._settings = settings
        self._object = object_.copy()

    def copy(self) -> FromMemoryObjectBuilder:
        builder = FromMemoryObjectBuilder(self._settings, self._object)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        object_geometry = geometry_provider.get_object_geometry()

        try:
            pixel_geometry = self._object.get_pixel_geometry()
        except ValueError:
            pixel_geometry = object_geometry.get_pixel_geometry()

        try:
            center = self._object.get_center()
        except ValueError:
            center = object_geometry.get_center()

        return Object(
            self._object.get_array(),
            pixel_geometry,
            center,
            self._object.layer_spacing_m,
        )


class FromFileObjectBuilder(ObjectBuilder):
    def __init__(
        self,
        settings: ObjectSettings,
        file_reader: ObjectFileReader,
    ) -> None:
        super().__init__(settings, 'from_file')
        self._settings = settings
        self._file_reader = file_reader

        self.file_path = settings.file_path.copy()
        self._add_parameter('file_path', self.file_path)

        self.file_type = settings.file_type.copy()
        self._add_parameter('file_type', self.file_type)

    def copy(self) -> FromFileObjectBuilder:
        builder = FromFileObjectBuilder(self._settings, self._file_reader)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        file_path = self.file_path.get_value()
        file_type = self.file_type.get_value()
        logger.debug(f'Reading "{file_path}" as "{file_type}"')

        try:
            object_from_file = self._file_reader.read(file_path)
        except Exception as exc:
            raise RuntimeError(f'Failed to read "{file_path}"') from exc

        object_geometry = geometry_provider.get_object_geometry()

        try:
            pixel_geometry = object_from_file.get_pixel_geometry()
        except ValueError:
            pixel_geometry = object_geometry.get_pixel_geometry()

        try:
            center = object_from_file.get_center()
        except ValueError:
            center = object_geometry.get_center()

        return Object(
            object_from_file.get_array(),
            pixel_geometry,
            center,
            object_from_file.layer_spacing_m,
        )



from __future__ import annotations
from abc import abstractmethod
from collections.abc import Iterable, Iterator, Sequence
import logging

import numpy

from ptychodus.api.geometry import AffineTransform
from ptychodus.api.parametric import ParameterGroup
from ptychodus.api.probe_positions import (
    ProbePosition,
    ProbePositionFileReader,
    ProbePositionSequence,
)
from ptychodus.api.probe_positions_gen import transform_probe_positions

from .settings import ProbePositionsSettings

logger = logging.getLogger(__name__)


class ProbePositionsBuilder(ParameterGroup):
    def __init__(
        self, rng: numpy.random.Generator, settings: ProbePositionsSettings, name: str
    ) -> None:
        super().__init__()
        self._rng = rng

        self._name = settings.builder.copy()
        self._name.set_value(name)
        self._add_parameter('name', self._name)

        self.affine00 = settings.affine00.copy()
        self._add_parameter('affine00', self.affine00)

        self.affine01 = settings.affine01.copy()
        self._add_parameter('affine01', self.affine01)

        self.affine02 = settings.affine02.copy()
        self._add_parameter('affine02', self.affine02)

        self.affine10 = settings.affine10.copy()
        self._add_parameter('affine10', self.affine10)

        self.affine11 = settings.affine11.copy()
        self._add_parameter('affine11', self.affine11)

        self.affine12 = settings.affine12.copy()
        self._add_parameter('affine12', self.affine12)

        self.jitter_radius_m = settings.jitter_radius_m.copy()
        self._add_parameter('jitter_radius_m', self.jitter_radius_m)

    def get_name(self) -> str:
        return self._name.get_value()

    @staticmethod
    def negate_x(preset: int) -> bool:
        return preset & 0x1 != 0x0

    @staticmethod
    def negate_y(preset: int) -> bool:
        return preset & 0x2 != 0x0

    @staticmethod
    def swap_xy(preset: int) -> bool:
        return preset & 0x4 != 0x0

    def labels_for_preset_transforms(self) -> Iterator[str]:
        for index in range(8):
            xp = '\u2212x' if self.negate_x(index) else '\u002bx'
            yp = '\u2212y' if self.negate_y(index) else '\u002by'
            fxy = f'{yp}, {xp}' if self.swap_xy(index) else f'{xp}, {yp}'
            yield f'(x, y) \u2192 ({fxy})'

    def assign_preset_transform(self, index: int) -> None:
        self.block_notifications(True)

        if self.swap_xy(index):
            self.affine00.set_value(0)
            self.affine01.set_value(-1 if self.negate_y(index) else +1)
            self.affine02.set_value(0)
            self.affine10.set_value(-1 if self.negate_x(index) else +1)
            self.affine11.set_value(0)
            self.affine12.set_value(0)
        else:
            self.affine00.set_value(-1 if self.negate_x(index) else +1)
            self.affine01.set_value(0)
            self.affine02.set_value(0)
            self.affine10.set_value(0)
            self.affine11.set_value(-1 if self.negate_y(index) else +1)
            self.affine12.set_value(0)

        self.block_notifications(False)

    def get_transform(self) -> AffineTransform:
        return AffineTransform(
            a00=self.affine00.get_value(),
            a01=self.affine01.get_value(),
            a02=self.affine02.get_value(),
            a10=self.affine10.get_value(),
            a11=self.affine11.get_value(),
            a12=self.affine12.get_value(),
        )

    def set_transform(self, transform: AffineTransform) -> None:
        self.block_notifications(True)

        self.affine00.set_value(transform.a00)
        self.affine01.set_value(transform.a01)
        self.affine02.set_value(transform.a02)

        self.affine10.set_value(transform.a10)
        self.affine11.set_value(transform.a11)
        self.affine12.set_value(transform.a12)

        self.block_notifications(False)

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

    @abstractmethod
    def copy(self) -> ProbePositionsBuilder:
        pass

    @abstractmethod
    def build(self) -> ProbePositionSequence:
        pass

    def _create_position_sequence(
        self, positions: Iterable[ProbePosition]
    ) -> ProbePositionSequence:
        transform = self.get_transform()
        jitter_radius_m = self.jitter_radius_m.get_value()
        rng = self._rng if jitter_radius_m > 0.0 else None
        return ProbePositionSequence(
            [*transform_probe_positions(positions, transform, rng, jitter_radius_m)]
        )


class FromMemoryProbePositionsBuilder(ProbePositionsBuilder):
    def __init__(
        self,
        rng: numpy.random.Generator,
        settings: ProbePositionsSettings,
        position_seq: Sequence[ProbePosition],
    ) -> None:
        super().__init__(rng, settings, 'from_memory')
        self._rng = rng
        self._settings = settings
        self._position_seq = ProbePositionSequence(position_seq)

        # set identity transformation
        self.assign_preset_transform(0)

    def copy(self) -> FromMemoryProbePositionsBuilder:
        builder = FromMemoryProbePositionsBuilder(self._rng, self._settings, self._position_seq)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self) -> ProbePositionSequence:
        return self._position_seq


class FromFileProbePositionsBuilder(ProbePositionsBuilder):
    def __init__(
        self,
        rng: numpy.random.Generator,
        settings: ProbePositionsSettings,
        file_reader: ProbePositionFileReader,
    ) -> None:
        super().__init__(rng, settings, 'from_file')
        self._rng = rng
        self._settings = settings
        self._file_reader = file_reader

        self.file_path = settings.file_path.copy()
        self._add_parameter('file_path', self.file_path)

        self.file_type = settings.file_type.copy()
        self._add_parameter('file_type', self.file_type)

    def copy(self) -> FromFileProbePositionsBuilder:
        builder = FromFileProbePositionsBuilder(self._rng, self._settings, self._file_reader)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self) -> ProbePositionSequence:
        file_path = self.file_path.get_value()
        file_type = self.file_type.get_value()
        logger.debug(f'Reading "{file_path}" as "{file_type}"')

        try:
            position_seq = self._file_reader.read(file_path)
        except Exception as exc:
            raise RuntimeError(f'Failed to read "{file_path}"') from exc

        return position_seq



from __future__ import annotations
from abc import abstractmethod
from collections.abc import Sequence
from enum import auto, IntEnum
import logging

import numpy

from ptychodus.api.parametric import ParameterGroup
from ptychodus.api.probe_gen import generate_coherent_probe_modes, generate_incoherent_probe_modes
from ptychodus.api.probe import (
    Probe,
    ProbeSequence,
    ProbeFileReader,
    ProbeGeometryProvider,
)

from .settings import ProbeSettings

logger = logging.getLogger(__name__)


class ProbeModeDecayType(IntEnum):
    NONE = auto()
    POLYNOMIAL = auto()
    EXPONENTIAL = auto()

    def get_weights(self, num_modes: int, decay_ratio: float) -> Sequence[float]:
        match self.value:
            case ProbeModeDecayType.EXPONENTIAL:
                b = 1.0 / decay_ratio
                return [b**-n for n in range(num_modes)]
            case ProbeModeDecayType.POLYNOMIAL:
                b = numpy.log(decay_ratio) / numpy.log(2.0)
                return [(n + 1) ** b for n in range(num_modes)]
            case _:
                return [1.0] + [0.0] * (num_modes - 1)


class ProbeSequenceBuilder(ParameterGroup):
    def __init__(self, settings: ProbeSettings, name: str) -> None:
        super().__init__()
        self._name = settings.builder.copy()
        self._name.set_value(name)
        self._add_parameter('name', self._name)

        self.num_incoherent_modes = settings.num_incoherent_modes.copy()
        self._add_parameter('num_incoherent_modes', self.num_incoherent_modes)

        self.orthogonalize_incoherent_modes = settings.orthogonalize_incoherent_modes.copy()
        self._add_parameter('orthogonalize_incoherent_modes', self.orthogonalize_incoherent_modes)

        self.incoherent_mode_decay_type = settings.incoherent_mode_decay_type.copy()
        self._add_parameter('incoherent_mode_decay_type', self.incoherent_mode_decay_type)

        self.incoherent_mode_decay_ratio = settings.incoherent_mode_decay_ratio.copy()
        self._add_parameter('incoherent_mode_decay_ratio', self.incoherent_mode_decay_ratio)

        self.num_coherent_modes = settings.num_coherent_modes.copy()
        self._add_parameter('num_coherent_modes', self.num_coherent_modes)

    def get_name(self) -> str:
        return self._name.get_value()

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

    @abstractmethod
    def copy(self) -> ProbeSequenceBuilder:
        pass

    @abstractmethod
    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        pass

    def _get_imode_weights(self) -> Sequence[float]:
        imode_decay_ratio = self.incoherent_mode_decay_ratio.get_value()
        imode_decay_type_text = self.incoherent_mode_decay_type.get_value()
        imode_decay_type = ProbeModeDecayType.NONE

        if imode_decay_ratio > 0.0:
            try:
                imode_decay_type = ProbeModeDecayType[imode_decay_type_text.upper()]
            except KeyError:
                logger.debug(f'Unknown probe mode decay type "{imode_decay_type_text}"')

        num_imodes = self.num_incoherent_modes.get_value()
        return imode_decay_type.get_weights(num_imodes, imode_decay_ratio)

    def _build_probe_modes(
        self, rng: numpy.random.Generator, probe: Probe, num_diffraction_patterns: int
    ) -> ProbeSequence:
        probe_with_imodes = generate_incoherent_probe_modes(
            rng,
            probe,
            self._get_imode_weights(),
            orthogonalize=self.orthogonalize_incoherent_modes.get_value(),
        )
        probe_seq = generate_coherent_probe_modes(
            rng,
            probe_with_imodes,
            num_cmodes=self.num_coherent_modes.get_value(),
            num_diffraction_patterns=num_diffraction_patterns,
        )
        array = probe_seq.get_array()
        logger.debug(f'Multimodal probe {array.shape=}')
        return probe_seq


class FromMemoryProbeBuilder(ProbeSequenceBuilder):
    def __init__(self, settings: ProbeSettings, probe: ProbeSequence) -> None:
        super().__init__(settings, 'from_memory')
        self._settings = settings
        self._probe = probe.copy()

    def copy(self) -> FromMemoryProbeBuilder:
        builder = FromMemoryProbeBuilder(self._settings, self._probe)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        probe_geometry = geometry_provider.get_probe_geometry()

        try:
            pixel_geometry = self._probe.get_pixel_geometry()
        except ValueError:
            pixel_geometry = probe_geometry.get_pixel_geometry()

        try:
            opr_weights = self._probe.get_opr_weights()
        except ValueError:
            opr_weights = None

        # TODO regrid probe as needed based on probe geometry from file/provider
        return ProbeSequence(
            self._probe.get_array(),
            opr_weights,
            pixel_geometry,
        )


class FromFileProbeBuilder(ProbeSequenceBuilder):
    def __init__(self, settings: ProbeSettings, file_reader: ProbeFileReader) -> None:
        super().__init__(settings, 'from_file')
        self._settings = settings
        self._file_reader = file_reader

        self.file_path = settings.file_path.copy()
        self._add_parameter('file_path', self.file_path)

        self.file_type = settings.file_type.copy()
        self._add_parameter('file_type', self.file_type)

    def copy(self) -> FromFileProbeBuilder:
        builder = FromFileProbeBuilder(self._settings, self._file_reader)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        file_path = self.file_path.get_value()
        file_type = self.file_type.get_value()
        logger.debug(f'Reading "{file_path}" as "{file_type}"')

        try:
            probe_from_file = self._file_reader.read(file_path)
        except Exception as exc:
            raise RuntimeError(f'Failed to read "{file_path}"') from exc

        probe_geometry = geometry_provider.get_probe_geometry()

        try:
            pixel_geometry = probe_from_file.get_pixel_geometry()
        except ValueError:
            pixel_geometry = probe_geometry.get_pixel_geometry()

        try:
            opr_weights = probe_from_file.get_opr_weights()
        except ValueError:
            opr_weights = None

        # TODO regrid probe as needed based on probe geometry from file/provider
        return ProbeSequence(
            probe_from_file.get_array(),
            opr_weights,
            pixel_geometry,
        )



from __future__ import annotations

from PyQt5.QtCore import QSize, Qt
from PyQt5.QtWidgets import (
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QListView,
    QSizePolicy,
    QSpacerItem,
    QStyle,
    QVBoxLayout,
    QWidget,
)


class SettingsView(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.list_view = QListView()
        self.button_box = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Open | QDialogButtonBox.StandardButton.Save
        )

        layout = QVBoxLayout()
        layout.addWidget(self.list_view)
        layout.addWidget(self.button_box)
        self.setLayout(layout)


class SyncProductToSettingsDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.icon_label = QLabel()
        self.prompt_label = QLabel('Update settings registry with product parameters?')
        self.product_label = QLabel('Product:')
        self.product_combo_box = QComboBox()
        self.button_box = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Yes | QDialogButtonBox.StandardButton.No
        )

        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        style = self.style()
        icon = style.standardIcon(QStyle.SP_MessageBoxQuestion)
        icon_size = style.pixelMetric(QStyle.PM_MessageBoxIconSize)
        pixmap = icon.pixmap(QSize(icon_size, icon_size))
        self.icon_label.setPixmap(pixmap)

        indent_spacer = QSpacerItem(7, QSizePolicy.Fixed, QSizePolicy.Fixed)

        layout = QGridLayout()
        layout.addWidget(self.icon_label, 0, 0, 2, 1, Qt.AlignTop)
        layout.addItem(indent_spacer, 0, 1, 2, 1)
        layout.addWidget(self.prompt_label, 0, 2, 1, 2)
        layout.addWidget(self.product_label, 1, 2, 1, 1)
        layout.addWidget(self.product_combo_box, 1, 3, 1, 1)
        layout.addWidget(self.button_box, 2, 0, 1, 4)
        self.setLayout(layout)

        self.setWindowTitle('Sync Data Product To Settings')



from __future__ import annotations
from collections.abc import Sequence
import logging

import numpy

from ptychodus.api.object import Object, ObjectGeometryProvider
from ptychodus.api.object_gen import generate_dead_leaves_object

from .builder import ObjectBuilder
from .settings import ObjectSettings

logger = logging.getLogger(__name__)


class DeadLeavesObjectBuilder(ObjectBuilder):
    def __init__(self, rng: numpy.random.Generator, settings: ObjectSettings) -> None:
        super().__init__(settings, 'dead_leaves')
        self._rng = rng
        self._settings = settings

        self.leaf_radius_lower_px = settings.leaf_radius_lower_px.copy()
        self._add_parameter('leaf_radius_lower_px', self.leaf_radius_lower_px)
        self.leaf_radius_upper_px = settings.leaf_radius_upper_px.copy()
        self._add_parameter('leaf_radius_upper_px', self.leaf_radius_upper_px)
        self.leaf_radius_power_law_exponent = settings.leaf_radius_power_law_exponent.copy()
        self._add_parameter('leaf_radius_power_law_exponent', self.leaf_radius_power_law_exponent)

        self.leaf_amplitude_lower = settings.leaf_amplitude_lower.copy()
        self._add_parameter('leaf_amplitude_lower', self.leaf_amplitude_lower)
        self.leaf_amplitude_upper = settings.leaf_amplitude_upper.copy()
        self._add_parameter('leaf_amplitude_upper', self.leaf_amplitude_upper)

        self.leaf_phase_lower_tr = settings.leaf_phase_lower_tr.copy()
        self._add_parameter('leaf_phase_lower_tr', self.leaf_phase_lower_tr)
        self.leaf_phase_upper_tr = settings.leaf_phase_upper_tr.copy()
        self._add_parameter('leaf_phase_upper_tr', self.leaf_phase_upper_tr)

    def copy(self) -> DeadLeavesObjectBuilder:
        builder = DeadLeavesObjectBuilder(self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        object_ = generate_dead_leaves_object(
            self._rng,
            geometry_provider.get_object_geometry(),
            leaf_radius_lower_px=self.leaf_radius_lower_px.get_value(),
            leaf_radius_upper_px=self.leaf_radius_upper_px.get_value(),
            leaf_radius_power_law_exponent=self.leaf_radius_power_law_exponent.get_value(),
            leaf_amplitude_lower=self.leaf_amplitude_lower.get_value(),
            leaf_amplitude_upper=self.leaf_amplitude_upper.get_value(),
            leaf_phase_lower_tr=self.leaf_phase_lower_tr.get_value(),
            leaf_phase_upper_tr=self.leaf_phase_upper_tr.get_value(),
        )
        return self._create_object(object_, layer_spacing_m)



from __future__ import annotations
import logging

import numpy

from ptychodus.api.geometry import ZernikeMonomial
from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider
from ptychodus.api.probe_gen import generate_zernike_probe, rescale_probe_intensity

from .builder import ProbeSequenceBuilder
from .settings import ProbeSettings

logger = logging.getLogger(__name__)


class ZernikeProbeBuilder(ProbeSequenceBuilder):
    def __init__(self, rng: numpy.random.Generator, settings: ProbeSettings) -> None:
        super().__init__(settings, 'zernike')
        self._rng = rng
        self._settings = settings
        self._polynomial: list[ZernikeMonomial] = list()
        self._order = 0

        self.diameter_m = settings.disk_diameter_m.copy()
        self._add_parameter('diameter_m', self.diameter_m)

        self.set_order(1)

    def copy(self) -> ZernikeProbeBuilder:
        builder = ZernikeProbeBuilder(self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        builder._polynomial = self._polynomial.copy()
        builder._order = self._order
        return builder

    def set_order(self, order: int) -> None:
        if order < 1:
            logger.warning('Order must be strictly positive!')
            return

        if self._order == order:
            return

        self._polynomial.clear()

        for radial_degree in range(order):
            for angular_frequency in range(-radial_degree, 1 + radial_degree, 2):
                monomial = ZernikeMonomial(1 + 0j, radial_degree, angular_frequency)
                self._polynomial.append(monomial)

        self._order = order
        self.notify_observers()

    def get_order(self) -> int:
        return self._order

    def set_coefficient(self, idx: int, value: complex) -> None:
        monomial = self._polynomial[idx]
        self._polynomial[idx] = ZernikeMonomial(
            value, monomial.radial_degree, monomial.angular_frequency
        )

    def get_monomial(self, idx: int) -> ZernikeMonomial:
        return self._polynomial[idx]

    def __len__(self) -> int:
        return len(self._polynomial)

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        probe = rescale_probe_intensity(
            generate_zernike_probe(
                geometry_provider.get_probe_geometry(),
                self._polynomial,
                radius_m=self.diameter_m.get_value() / 2.0,
            ),
            geometry_provider.probe_photon_count,
        )
        return self._build_probe_modes(self._rng, probe, geometry_provider.num_scan_points)



from collections.abc import Iterable, Iterator
from itertools import pairwise

import numpy

from .geometry import AffineTransform
from .probe_positions import ProbePosition, ScanGeometry


def generate_cartesian_probe_positions(
    num_points_x: int,
    num_points_y: int,
    step_size_x: float,
    step_size_y: float,
    *,
    snake: bool = False,
    stagger: bool = False,
) -> Iterator[ProbePosition]:
    for index in range(num_points_x * num_points_y):
        y, x = divmod(index, num_points_x)

        if snake:
            if y & 1:
                x = num_points_x - 1 - x

        cx = (num_points_x - 1) / 2
        cy = (num_points_y - 1) / 2

        xf = (x - cx) * step_size_x
        yf = (y - cy) * step_size_y

        if stagger:
            if y & 1:
                xf += step_size_x / 4
            else:
                xf -= step_size_x / 4

        yield ProbePosition(
            index=index,
            coordinate_x_m=xf,
            coordinate_y_m=yf,
        )


def generate_concentric_probe_positions(
    radial_step_size_m: float, num_shells: int, num_points_1st_shell: int
) -> Iterator[ProbePosition]:
    """https://doi.org/10.1088/1367-2630/12/3/035017"""
    triangle = (num_shells * (num_shells + 1)) // 2
    num_points = triangle * num_points_1st_shell

    for index in range(num_points):
        triangle = index // num_points_1st_shell
        shell_index = int((1 + numpy.sqrt(1 + 8 * triangle)) / 2) - 1  # see OEIS A002024
        shell_triangle = (shell_index * (shell_index + 1)) // 2
        first_index_in_shell = num_points_1st_shell * shell_triangle
        point_index_in_shell = index - first_index_in_shell

        radius_m = radial_step_size_m * (shell_index + 1)
        num_points_in_shell = num_points_1st_shell * (shell_index + 1)
        theta_rad = 2 * numpy.pi * point_index_in_shell / num_points_in_shell

        yield ProbePosition(
            index=index,
            coordinate_x_m=radius_m * numpy.cos(theta_rad),
            coordinate_y_m=radius_m * numpy.sin(theta_rad),
        )


def generate_lissajous_probe_positions(
    num_points: int,
    amplitude_x_m: float,
    amplitude_y_m: float,
    angular_step_x_tr: float,
    angular_step_y_tr: float,
    angular_shift_tr: float,
) -> Iterator[ProbePosition]:
    for index in range(num_points):
        two_pi = 2 * numpy.pi
        theta_x = two_pi * angular_step_x_tr * index + angular_shift_tr
        theta_y = two_pi * angular_step_y_tr * index

        yield ProbePosition(
            index=index,
            coordinate_x_m=amplitude_x_m * numpy.sin(theta_x),
            coordinate_y_m=amplitude_y_m * numpy.sin(theta_y),
        )


def generate_spiral_probe_positions(
    num_points: int, radius_scalar_m: float
) -> Iterator[ProbePosition]:
    """https://doi.org/10.1364/OE.22.012634"""
    for index in range(num_points):
        radius_m = radius_scalar_m * numpy.sqrt(index)
        divergence_angle_rad = (3.0 - numpy.sqrt(5)) * numpy.pi
        theta_rad = divergence_angle_rad * index

        yield ProbePosition(
            index=index,
            coordinate_x_m=radius_m * numpy.cos(theta_rad),
            coordinate_y_m=radius_m * numpy.sin(theta_rad),
        )


def transform_probe_positions(
    positions: Iterable[ProbePosition],
    transform: AffineTransform,
    rng: numpy.random.Generator | None = None,
    jitter_radius_m: float = 0.0,
) -> Iterator[ProbePosition]:
    for position in positions:
        x_m, y_m = transform(position.coordinate_x_m, position.coordinate_y_m)

        if rng is not None:
            angle_rad = 2 * numpy.pi * rng.uniform()
            radius_m = jitter_radius_m * numpy.sqrt(rng.uniform())

            x_m += radius_m * numpy.cos(angle_rad)
            y_m += radius_m * numpy.sin(angle_rad)

        yield ProbePosition(
            index=position.index,
            coordinate_x_m=x_m,
            coordinate_y_m=y_m,
        )


def calculate_scan_geometry(positions: Iterable[ProbePosition]) -> ScanGeometry | None:
    minimum_x_m = +numpy.inf
    maximum_x_m = -numpy.inf
    minimum_y_m = +numpy.inf
    maximum_y_m = -numpy.inf
    length_m = 0.0

    for point in positions:
        if point.coordinate_x_m < minimum_x_m:
            minimum_x_m = point.coordinate_x_m

        if maximum_x_m < point.coordinate_x_m:
            maximum_x_m = point.coordinate_x_m

        if point.coordinate_y_m < minimum_y_m:
            minimum_y_m = point.coordinate_y_m

        if maximum_y_m < point.coordinate_y_m:
            maximum_y_m = point.coordinate_y_m

    is_empty_x = maximum_x_m < minimum_x_m
    is_empty_y = maximum_y_m < minimum_y_m

    if is_empty_x or is_empty_y:
        return None

    for point_l, point_r in pairwise(positions):
        dx = point_r.coordinate_x_m - point_l.coordinate_x_m
        dy = point_r.coordinate_y_m - point_l.coordinate_y_m
        length_m += numpy.hypot(dx, dy)

    return ScanGeometry(
        minimum_x_m=minimum_x_m,
        maximum_x_m=maximum_x_m,
        minimum_y_m=minimum_y_m,
        maximum_y_m=maximum_y_m,
        length_m=length_m,
    )



"""Tests for _generate_simplex_noise and related helpers in ptychodus.api.object_gen.

Key behaviors verified:
  - Correct output shape and dtype
  - No NaN / Inf values
  - Reproducibility with a fixed RNG seed
  - Non-reproducibility with different seeds
  - Near-zero sample mean (gradient symmetry implies E[noise] = 0)
  - Bounded output amplitude
  - Spatial smoothness compared to white noise
  - Correlation length proportional to grid_scale_px
  - Dominant spatial frequency inversely proportional to grid_scale_px
  - Coordinate-mapping roundtrip consistency (simplex <-> Cartesian)
  - Visual outputs saved to tmp_path for manual inspection

Note on kernel support
----------------------
The kernel is `max(0, 0.5 - d^2/grid_scale_px^2)^4` where `d` is the
pixel-space distance from a vertex.  The fixed threshold 0.5 is identical
to the standard simplex-noise convention (Gustavson 2012), normalised so
that all three vertices of each simplex cell contribute at every interior
point regardless of `grid_scale_px`.
"""

import matplotlib

matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy
import numpy.testing
import pytest

from ptychodus.api.object_gen import (
    _calculate_vertex_noise_contribution,
    _generate_simplex_noise,
    _map_cartesian_to_simplex,
    _map_simplex_to_cartesian,
)


# ---------------------------------------------------------------------------
# Helpers shared by multiple tests
# ---------------------------------------------------------------------------


def _rng(seed: int = 42) -> numpy.random.Generator:
    return numpy.random.default_rng(seed)


def _radial_power_spectrum(
    image: numpy.ndarray, num_bins: int = 64
) -> tuple[numpy.ndarray, numpy.ndarray]:
    """Radially averaged power spectrum; returns (frequencies, power) in cycles/pixel."""
    h, w = image.shape
    power = numpy.abs(numpy.fft.fftshift(numpy.fft.fft2(image))) ** 2
    fy = numpy.fft.fftshift(numpy.fft.fftfreq(h))
    fx = numpy.fft.fftshift(numpy.fft.fftfreq(w))
    FX, FY = numpy.meshgrid(fx, fy)  # noqa: N806
    R = numpy.hypot(FX, FY)  # noqa: N806
    r_max = min(fy.max(), fx.max())
    edges = numpy.linspace(0.0, r_max, num_bins + 1)
    centers = 0.5 * (edges[:-1] + edges[1:])
    avg_power = numpy.zeros(num_bins)
    for k in range(num_bins):
        mask = (R >= edges[k]) & (R < edges[k + 1])
        if mask.any():
            avg_power[k] = power[mask].mean()
    return centers, avg_power


def _correlation_length(image: numpy.ndarray) -> float:
    """Estimate 1-D correlation length (pixels) from the row-averaged autocorrelation."""
    row = image.mean(axis=0)
    row = row - row.mean()
    if row.std() == 0.0:
        return 0.0
    n = len(row)
    acf_full = numpy.fft.irfft(numpy.abs(numpy.fft.rfft(row, n=2 * n)) ** 2)[:n]
    acf = acf_full / acf_full[0]
    crossings = numpy.where(acf < numpy.exp(-1.0))[0]
    return float(crossings[0]) if len(crossings) else float(n)


# ---------------------------------------------------------------------------
# Shape and dtype
# ---------------------------------------------------------------------------


class TestOutputShapeAndDtype:
    def test_square_image(self) -> None:
        noise = _generate_simplex_noise(_rng(), 64, 64, 4.0)
        assert noise.shape == (64, 64)

    def test_rectangular_wide(self) -> None:
        noise = _generate_simplex_noise(_rng(), 80, 40, 4.0)
        assert noise.shape == (40, 80)

    def test_rectangular_tall(self) -> None:
        noise = _generate_simplex_noise(_rng(), 40, 80, 4.0)
        assert noise.shape == (80, 40)

    def test_dtype_is_floating(self) -> None:
        noise = _generate_simplex_noise(_rng(), 32, 32, 4.0)
        assert numpy.issubdtype(noise.dtype, numpy.floating)

    def test_no_nans(self) -> None:
        noise = _generate_simplex_noise(_rng(), 64, 64, 4.0)
        assert not numpy.any(numpy.isnan(noise))

    def test_no_infs(self) -> None:
        noise = _generate_simplex_noise(_rng(), 64, 64, 4.0)
        assert not numpy.any(numpy.isinf(noise))


# ---------------------------------------------------------------------------
# Reproducibility
# ---------------------------------------------------------------------------


class TestReproducibility:
    def test_same_seed_identical_output(self) -> None:
        noise1 = _generate_simplex_noise(_rng(7), 64, 64, 4.0)
        noise2 = _generate_simplex_noise(_rng(7), 64, 64, 4.0)
        numpy.testing.assert_array_equal(noise1, noise2)

    def test_different_seeds_different_output(self) -> None:
        noise1 = _generate_simplex_noise(_rng(7), 64, 64, 4.0)
        noise2 = _generate_simplex_noise(_rng(99), 64, 64, 4.0)
        assert not numpy.array_equal(noise1, noise2)

    def test_changing_grid_scale_changes_output(self) -> None:
        noise1 = _generate_simplex_noise(_rng(0), 64, 64, 4.0)
        noise2 = _generate_simplex_noise(_rng(0), 64, 64, 8.0)
        assert not numpy.allclose(noise1, noise2)


# ---------------------------------------------------------------------------
# Statistical properties
# ---------------------------------------------------------------------------


class TestStatisticalProperties:
    """Gradient directions are drawn uniformly from [0, 2π], so E[noise] = 0
    by symmetry, and amplitude is bounded by the kernel maximum times 70.
    """

    def test_nonzero_variance(self) -> None:
        noise = _generate_simplex_noise(_rng(), 64, 64, 8.0)
        assert noise.std() > 0.0, 'Expected non-trivial noise (all-zero output)'

    def test_mean_near_zero(self) -> None:
        """Sample mean should be small relative to std over a large image."""
        noise = _generate_simplex_noise(_rng(), 256, 256, 8.0)
        # Relative threshold: mean < 30 % of std
        assert abs(noise.mean()) < 0.3 * noise.std(), (
            f'Mean ({noise.mean():.4f}) is large relative to std ({noise.std():.4f}); '
            'expected near-zero mean from symmetric random gradients.'
        )

    def test_values_bounded(self) -> None:
        """Noise values should be bounded for a small grid scale."""
        noise = _generate_simplex_noise(_rng(), 128, 128, 1.0)
        assert noise.max() < 10.0
        assert noise.min() > -10.0

    def test_distribution_roughly_symmetric(self) -> None:
        """Amplitude distribution should be approximately symmetric (low skewness)."""
        noise = _generate_simplex_noise(_rng(1), 256, 256, 8.0)
        centered = noise - noise.mean()
        var = numpy.mean(centered**2)
        if var > 0:
            skewness = float(numpy.mean(centered**3) / var**1.5)
            assert abs(skewness) < 1.0, f'Skewness {skewness:.3f} too large for noise'


# ---------------------------------------------------------------------------
# Spatial properties
# ---------------------------------------------------------------------------


class TestSpatialProperties:
    """Simplex noise should be spatially smooth with a characteristic scale
    set by grid_scale_px.
    """

    def test_smoothness_compared_to_white_noise(self) -> None:
        """Normalized gradient magnitude must be well below the white-noise baseline.

        For i.i.d. white noise: RMS(grad) / std(noise) ≈ sqrt(2*2) ≈ 2.
        For smooth noise with correlation length L:
            RMS(grad) / std(noise) ≈ sqrt(2) / L.
        With grid_scale_px = 8 we expect the ratio to be << 1.
        """
        noise = _generate_simplex_noise(_rng(), 128, 128, 8.0)
        gy = numpy.diff(noise, axis=0)
        gx = numpy.diff(noise, axis=1)
        rms_grad = numpy.sqrt(numpy.mean(gy**2) + numpy.mean(gx**2))
        normalized = rms_grad / noise.std()
        assert normalized < 1.0, (
            f'Normalized gradient magnitude {normalized:.3f} is too large; '
            'expected smoother than white noise (baseline ≈ 2).'
        )

    def test_correlation_length_scales_with_grid_scale(self) -> None:
        """Larger grid_scale_px → longer spatial correlation length."""
        scale_small, scale_large = 4.0, 16.0
        noise_small = _generate_simplex_noise(_rng(3), 256, 256, scale_small)
        noise_large = _generate_simplex_noise(_rng(3), 256, 256, scale_large)
        len_small = _correlation_length(noise_small)
        len_large = _correlation_length(noise_large)
        assert len_large > len_small, (
            f'Expected longer correlation length for scale={scale_large} '
            f'({len_large:.1f} px) vs scale={scale_small} ({len_small:.1f} px).'
        )

    def test_dominant_frequency_decreases_with_grid_scale(self) -> None:
        """Peak spatial frequency should decrease as grid_scale_px increases."""
        scale_small, scale_large = 4.0, 16.0
        freqs_s, power_s = _radial_power_spectrum(
            _generate_simplex_noise(_rng(1), 256, 256, scale_small)
        )
        freqs_l, power_l = _radial_power_spectrum(
            _generate_simplex_noise(_rng(1), 256, 256, scale_large)
        )
        # Skip DC bin (index 0)
        peak_small = freqs_s[1:][numpy.argmax(power_s[1:])]
        peak_large = freqs_l[1:][numpy.argmax(power_l[1:])]
        assert peak_small > peak_large, (
            f'Peak frequency should decrease with larger scale: '
            f'{peak_small:.4f} (scale={scale_small}) vs {peak_large:.4f} (scale={scale_large}).'
        )

    def test_characteristic_band_contains_most_power(self) -> None:
        """Most spectral power should fall near the characteristic scale 1/grid_scale_px."""
        scale = 8.0
        noise = _generate_simplex_noise(_rng(5), 256, 256, scale)
        freqs, power = _radial_power_spectrum(noise)
        f_char = 1.0 / scale
        power_mid = power[(freqs >= f_char / 4) & (freqs <= 4 * f_char)].sum()
        total = power.sum()
        if total > 0:
            assert power_mid / total > 0.3, (
                f'Only {100 * power_mid / total:.1f}% of power in characteristic band '
                f'[{f_char / 4:.4f}, {4 * f_char:.4f}] cyc/px; expected ≥ 30%.'
            )

    def test_correlation_length_is_proportional_to_grid_scale(self) -> None:
        """Correlation length should grow roughly linearly with grid_scale_px."""
        scales = [4.0, 8.0, 16.0]
        lengths = [
            _correlation_length(_generate_simplex_noise(_rng(0), 256, 256, s)) for s in scales
        ]
        # Each doubling of scale should increase the correlation length
        assert lengths[1] > lengths[0], (
            f'Doubling scale from {scales[0]} to {scales[1]} should increase '
            f'correlation length: {lengths[0]:.1f} → {lengths[1]:.1f} px.'
        )
        assert lengths[2] > lengths[1], (
            f'Doubling scale from {scales[1]} to {scales[2]} should increase '
            f'correlation length: {lengths[1]:.1f} → {lengths[2]:.1f} px.'
        )


# ---------------------------------------------------------------------------
# Coordinate-mapping roundtrip
# ---------------------------------------------------------------------------


class TestCoordinateMappings:
    """_map_simplex_to_cartesian and _map_cartesian_to_simplex must be inverses."""

    def test_simplex_to_cartesian_and_back(self) -> None:
        rng = _rng()
        xx = rng.uniform(0.0, 100.0, (32, 32))
        yy = rng.uniform(0.0, 100.0, (32, 32))
        scale = 10.0
        ii, jj = _map_simplex_to_cartesian(xx, yy, scale)
        xx_rt, yy_rt = _map_cartesian_to_simplex(ii, jj, scale)
        numpy.testing.assert_allclose(xx_rt, xx, rtol=1e-10, atol=1e-12)
        numpy.testing.assert_allclose(yy_rt, yy, rtol=1e-10, atol=1e-12)

    def test_cartesian_to_simplex_and_back(self) -> None:
        rng = _rng()
        ii = rng.uniform(-5.0, 5.0, (32, 32))
        jj = rng.uniform(-5.0, 5.0, (32, 32))
        scale = 7.0
        xx, yy = _map_cartesian_to_simplex(ii, jj, scale)
        ii_rt, jj_rt = _map_simplex_to_cartesian(xx, yy, scale)
        numpy.testing.assert_allclose(ii_rt, ii, rtol=1e-10, atol=1e-12)
        numpy.testing.assert_allclose(jj_rt, jj, rtol=1e-10, atol=1e-12)

    def test_origin_maps_to_origin(self) -> None:
        xx = numpy.zeros((1, 1))
        yy = numpy.zeros((1, 1))
        ii, jj = _map_simplex_to_cartesian(xx, yy, 5.0)
        numpy.testing.assert_allclose(ii, 0.0, atol=1e-12)
        numpy.testing.assert_allclose(jj, 0.0, atol=1e-12)

    def test_simplex_indices_scale_inversely_with_grid_scale(self) -> None:
        """Doubling grid_scale_px should halve the simplex indices ii, jj."""
        xx = numpy.full((4, 4), 10.0)
        yy = numpy.full((4, 4), 6.0)
        ii1, jj1 = _map_simplex_to_cartesian(xx, yy, grid_scale_px=5.0)
        ii2, jj2 = _map_simplex_to_cartesian(xx, yy, grid_scale_px=10.0)
        numpy.testing.assert_allclose(ii2, ii1 / 2.0, rtol=1e-10)
        numpy.testing.assert_allclose(jj2, jj1 / 2.0, rtol=1e-10)

    def test_vertex_contribution_zero_at_vertex(self) -> None:
        """Kernel × gradient dot product is zero at the vertex itself (displacement = 0)."""
        width, height = 16, 16
        yy, xx = numpy.mgrid[:height, :width].astype(float)
        vertex_i = numpy.zeros((height, width), dtype=int)
        vertex_j = numpy.zeros((height, width), dtype=int)
        rng = _rng()
        angle = 2 * numpy.pi * rng.uniform(size=(height, width))
        grad_x = numpy.cos(angle)
        grad_y = numpy.sin(angle)
        # Place a vertex exactly at pixel (0, 0); all other pixels have nonzero displacement
        scale = 4.0
        contrib = _calculate_vertex_noise_contribution(
            xx, yy, vertex_i, vertex_j, grad_x, grad_y, scale
        )
        # At pixel (0,0) the displacement is exactly zero → contribution = 0
        assert contrib[0, 0] == pytest.approx(0.0, abs=1e-12)


# ---------------------------------------------------------------------------
# Visual tests  (outputs saved to pytest's tmp_path for manual inspection)
# ---------------------------------------------------------------------------


class TestVisualOutput:
    def test_single_noise_image(self, tmp_path) -> None:  # type: ignore[no-untyped-def]
        """Save a grayscale simplex noise image."""
        scale = 16.0
        noise = _generate_simplex_noise(_rng(), 256, 256, scale)
        vmax = float(numpy.abs(noise).max()) or 1.0

        fig, ax = plt.subplots(figsize=(5, 5))
        im = ax.imshow(noise, cmap='gray', vmin=-vmax, vmax=vmax, interpolation='nearest')
        fig.colorbar(im, ax=ax, shrink=0.8)
        ax.set_title(f'Simplex Noise  scale={scale:.0f} px  std={noise.std():.3f}')
        ax.axis('off')
        fig.tight_layout()
        out = tmp_path / 'simplex_noise_single.png'
        fig.savefig(out, dpi=100)
        plt.close(fig)
        assert out.exists()

    def test_multi_scale_comparison(self, tmp_path) -> None:  # type: ignore[no-untyped-def]
        """Compare simplex noise texture at four grid scales side by side."""
        scales = [4.0, 8.0, 16.0, 32.0]
        fig, axes = plt.subplots(1, len(scales), figsize=(16, 4))
        for ax, scale in zip(axes, scales):
            noise = _generate_simplex_noise(_rng(0), 128, 128, scale)
            vmax = float(numpy.abs(noise).max()) or 1.0
            ax.imshow(noise, cmap='seismic', vmin=-vmax, vmax=vmax, interpolation='nearest')
            ax.set_title(f'scale={scale:.0f} px\nstd={noise.std():.3f}')
            ax.axis('off')
        fig.suptitle('Simplex Noise at Different Grid Scales')
        fig.tight_layout()
        out = tmp_path / 'simplex_noise_scales.png'
        fig.savefig(out, dpi=100)
        plt.close(fig)
        assert out.exists()

    def test_power_spectrum_plot(self, tmp_path) -> None:  # type: ignore[no-untyped-def]
        """Plot radial power spectra for several scales; characteristic frequency marked."""
        scales = [4.0, 8.0, 16.0]
        fig, ax = plt.subplots(figsize=(7, 4))
        for scale in scales:
            noise = _generate_simplex_noise(_rng(0), 256, 256, scale)
            freqs, power = _radial_power_spectrum(noise)
            ax.semilogy(freqs, power + 1e-30, label=f'scale={scale:.0f} px')
            ax.axvline(1.0 / scale, linestyle='--', alpha=0.4, color='gray')
        ax.set_xlabel('Spatial frequency (cycles/pixel)')
        ax.set_ylabel('Mean power (log scale)')
        ax.set_title('Radial Power Spectrum of Simplex Noise\n(dashed lines = 1/scale)')
        ax.legend()
        fig.tight_layout()
        out = tmp_path / 'simplex_noise_spectrum.png'
        fig.savefig(out, dpi=100)
        plt.close(fig)
        assert out.exists()

    def test_autocorrelation_plot(self, tmp_path) -> None:  # type: ignore[no-untyped-def]
        """Plot 1-D autocorrelation functions; verify 1/e crossing scales with grid_scale."""
        scales = [4.0, 8.0, 16.0]
        fig, ax = plt.subplots(figsize=(7, 4))
        crossings: list[tuple[float, float]] = []
        for scale in scales:
            noise = _generate_simplex_noise(_rng(0), 256, 256, scale)
            row = noise.mean(axis=0)
            row = row - row.mean()
            n = len(row)
            acf_full = numpy.fft.irfft(numpy.abs(numpy.fft.rfft(row, n=2 * n)) ** 2)[:n]
            acf = acf_full / acf_full[0] if acf_full[0] > 0 else acf_full
            lags = numpy.arange(n)
            ax.plot(lags[: n // 2], acf[: n // 2], label=f'scale={scale:.0f} px')
            idx = numpy.where(acf < numpy.exp(-1.0))[0]
            crossing = float(idx[0]) if len(idx) else float(n)
            crossings.append((scale, crossing))

        ax.axhline(numpy.exp(-1.0), linestyle=':', color='black', label='1/e')
        ax.set_xlabel('Lag (pixels)')
        ax.set_ylabel('Normalised autocorrelation')
        ax.set_title('1-D Autocorrelation of Simplex Noise')
        ax.set_xlim(0, 80)
        ax.legend()
        fig.tight_layout()
        out = tmp_path / 'simplex_noise_autocorrelation.png'
        fig.savefig(out, dpi=100)
        plt.close(fig)
        assert out.exists()

        # Quantitative check embedded in the visual test:
        # 1/e crossing should increase with scale
        for (s1, c1), (s2, c2) in zip(crossings[:-1], crossings[1:]):
            assert c2 > c1, (
                f'1/e crossing should increase from scale={s1} ({c1:.1f} px) '
                f'to scale={s2} ({c2:.1f} px).'
            )



from collections.abc import Iterable, Sequence
from typing import Final
import logging

from scipy.fft import fftfreq, ifft2
from scipy.interpolate import griddata
from scipy.ndimage import gaussian_filter
import numpy

from ._phase_unwrapper import PhaseUnwrapper
from .common import IntegerArrayType, RealArrayType, lerp
from .object import Object, ObjectGeometry
from .probe_positions import ProbePosition
from .reconstructor import AssembledDiffractionData

logger = logging.getLogger(__name__)


def generate_stxm_object(
    geometry: ObjectGeometry,
    assembled_data: AssembledDiffractionData,
    probe_positions: Iterable[ProbePosition],
) -> Object:
    coordinates_px: list[float] = list()
    values: list[float] = list()

    assembled_indexes = assembled_data.get_indexes().tolist()
    assembled_pattern_counts = assembled_data.get_pattern_counts().tolist()
    pattern_counts_lut = dict(zip(assembled_indexes, assembled_pattern_counts))

    for scan_point in probe_positions:
        try:
            value = pattern_counts_lut[scan_point.index]
        except KeyError:
            logger.debug(f'Skipping missing scan point index={scan_point.index}!')
        else:
            object_point = geometry.map_coordinates_probe_to_object(scan_point)
            coordinates_px.append(object_point.coordinate_y_px)
            coordinates_px.append(object_point.coordinate_x_px)
            values.append(value)

    points = numpy.reshape(coordinates_px, (-1, 2))
    YY, XX = numpy.mgrid[: geometry.height_px, : geometry.width_px]  # noqa: N806
    query_points = numpy.transpose((YY.flat, XX.flat))
    intensity = griddata(points, values, query_points, method='linear', fill_value=0.0).reshape(
        XX.shape
    )

    return Object(
        array=numpy.sqrt(intensity[numpy.newaxis, :, :]).astype('complex'),
        pixel_geometry=geometry.get_pixel_geometry(),
        center=geometry.get_center(),
    )


def generate_random_object(
    rng: numpy.random.Generator,
    geometry: ObjectGeometry,
    *,
    amplitude_mean: float,
    amplitude_deviation: float,
    phase_mean: float,
    phase_deviation_tr: float,
    blur_deviation_px: float,
) -> Object:
    object_shape = (1, geometry.height_px, geometry.width_px)

    amplitude = rng.normal(
        amplitude_mean,
        amplitude_deviation,
        object_shape,
    )
    phase = rng.normal(
        0.0,
        phase_deviation_tr,
        object_shape,
    )

    if blur_deviation_px > 0.0:
        # TODO account for pixel aspect ratio in blur calculation
        amplitude = gaussian_filter(amplitude, sigma=blur_deviation_px)
        phase = gaussian_filter(phase, sigma=blur_deviation_px)

    array = amplitude * numpy.exp(2j * numpy.pi * phase)

    return Object(
        array=array.astype('complex'),
        pixel_geometry=geometry.get_pixel_geometry(),
        center=geometry.get_center(),
    )


def generate_gaussian_random_field_object(
    rng: numpy.random.Generator, geometry: ObjectGeometry, *, correlation_length_px: float
) -> Object:
    """Generate a complex-valued Gaussian random field via spectral synthesis."""
    # white noise in Fourier space
    object_shape = (1, geometry.height_px, geometry.width_px)
    noise = rng.normal(size=object_shape) + 1j * rng.normal(size=object_shape)

    # frequency grid
    kx = fftfreq(geometry.width_px)
    ky = fftfreq(geometry.height_px)
    KX, KY = numpy.meshgrid(kx, ky)  # noqa: N806
    K2 = numpy.square(KX) + numpy.square(KY)  # noqa: N806

    # power spectrum: Gaussian envelope to control correlation length
    power_spectrum = numpy.exp(-0.5 * (2 * numpy.pi * correlation_length_px) ** 2 * K2)
    power_spectrum[0, 0] = 0  # zero mean

    # shape noise with power spectrum and back-transform
    field = numpy.asarray(ifft2(noise * numpy.sqrt(power_spectrum)))

    return Object(
        array=field,
        pixel_geometry=geometry.get_pixel_geometry(),
        center=geometry.get_center(),
    )


def _map_simplex_to_cartesian(
    xx: RealArrayType, yy: RealArrayType, grid_scale_px: float
) -> tuple[RealArrayType, RealArrayType]:
    SQRT3: Final[float] = numpy.sqrt(3)  # noqa: N806
    SQRT6: Final[float] = numpy.sqrt(6)  # noqa: N806

    c = 1 / (grid_scale_px * SQRT6)
    zs = xx + yy
    zd = xx - yy

    ii = c * (+zs + SQRT3 * zd)
    jj = c * (-zs + SQRT3 * zd)

    return ii, jj


def _map_cartesian_to_simplex(
    ii: RealArrayType, jj: RealArrayType, grid_scale_px: float
) -> tuple[RealArrayType, RealArrayType]:
    SQRT3: Final[float] = numpy.sqrt(3)  # noqa: N806
    SQRT8: Final[float] = numpy.sqrt(8)  # noqa: N806

    c = grid_scale_px / SQRT8
    ks = ii + jj
    kd = ii - jj

    xx = c * (ks + SQRT3 * kd)
    yy = c * (-ks + SQRT3 * kd)

    return xx, yy


def _calculate_vertex_noise_contribution(
    xx: RealArrayType,
    yy: RealArrayType,
    vertex_i: IntegerArrayType,
    vertex_j: IntegerArrayType,
    vertex_grad_x: RealArrayType,
    vertex_grad_y: RealArrayType,
    grid_scale_px: float,
) -> RealArrayType:
    vertex_x, vertex_y = _map_cartesian_to_simplex(
        vertex_i.astype(float), vertex_j.astype(float), grid_scale_px
    )
    displacement_x = (xx - vertex_x) / grid_scale_px
    displacement_y = (yy - vertex_y) / grid_scale_px
    distancesq_normalized = numpy.square(displacement_x) + numpy.square(displacement_y)
    kernel = numpy.maximum(0.0, 0.5 - distancesq_normalized) ** 4
    grad_dir_x = vertex_grad_x[vertex_j, vertex_i] * displacement_x
    grad_dir_y = vertex_grad_y[vertex_j, vertex_i] * displacement_y
    return kernel * (grad_dir_x + grad_dir_y)


def _generate_simplex_noise(
    rng: numpy.random.Generator,
    width_px: int,
    height_px: int,
    grid_scale_px: float,
) -> RealArrayType:
    # generate coordinate grid
    yy, xx = numpy.mgrid[:height_px, :width_px]

    # generate random direction vectors
    geometry_coefficient = (1 + numpy.sqrt(3)) / (grid_scale_px * numpy.sqrt(6))
    grad_shape_x = numpy.ceil(geometry_coefficient * width_px).astype(int) + 1
    grad_shape_y = numpy.ceil(geometry_coefficient * height_px).astype(int) + 1
    grad_shape = grad_shape_y, grad_shape_x
    logger.debug(f'{geometry_coefficient=} {grad_shape=}')
    angle_rad = 2 * numpy.pi * rng.uniform(size=grad_shape)
    vertex_grad_x = numpy.cos(angle_rad)
    vertex_grad_y = numpy.sin(angle_rad)

    # locate containing cell
    ii, jj = _map_simplex_to_cartesian(xx, yy, grid_scale_px)
    cell_origin_i = numpy.floor(ii).astype(int)
    cell_origin_j = numpy.floor(jj).astype(int)
    cell_origin_x, cell_origin_y = _map_cartesian_to_simplex(
        cell_origin_i.astype(float), cell_origin_j.astype(float), grid_scale_px
    )

    # vertex indexes
    is_lower_triangle = (ii - cell_origin_i) > (jj - cell_origin_j)
    vertex0_i = cell_origin_i
    vertex0_j = cell_origin_j
    vertex1_i = cell_origin_i + numpy.where(is_lower_triangle, 1, 0)
    vertex1_j = cell_origin_j + numpy.where(is_lower_triangle, 0, 1)
    vertex2_i = cell_origin_i + 1
    vertex2_j = cell_origin_j + 1

    # vertex noise contributions
    noise0 = _calculate_vertex_noise_contribution(
        xx, yy, vertex0_i, vertex0_j, vertex_grad_x, vertex_grad_y, grid_scale_px
    )
    noise1 = _calculate_vertex_noise_contribution(
        xx, yy, vertex1_i, vertex1_j, vertex_grad_x, vertex_grad_y, grid_scale_px
    )
    noise2 = _calculate_vertex_noise_contribution(
        xx, yy, vertex2_i, vertex2_j, vertex_grad_x, vertex_grad_y, grid_scale_px
    )

    # accumulate vertex contributions to noise
    noise = noise0 + noise1 + noise2

    noise_max = noise.max()
    noise_min = noise.min()
    return 2 * (noise - noise_min) / (noise_max - noise_min) - 1


def generate_simplex_noise_object(
    rng: numpy.random.Generator,
    geometry: ObjectGeometry,
    *,
    grid_scale_px: float = 30.0,
) -> Object:
    pixel_geometry = geometry.get_pixel_geometry()

    if not pixel_geometry.is_square:
        raise ValueError('Non-square pixels are unsupported!')

    if grid_scale_px <= 0.0:
        raise ValueError('Grid scale must be strictly positive!')

    re = _generate_simplex_noise(
        rng=rng,
        width_px=geometry.width_px,
        height_px=geometry.height_px,
        grid_scale_px=grid_scale_px,
    )
    im = _generate_simplex_noise(
        rng=rng,
        width_px=geometry.width_px,
        height_px=geometry.height_px,
        grid_scale_px=grid_scale_px,
    )
    array = re + 1j * im

    return Object(
        array=array,
        pixel_geometry=geometry.get_pixel_geometry(),
        center=geometry.get_center(),
    )


def generate_fractal_noise_object(
    rng: numpy.random.Generator,
    geometry: ObjectGeometry,
    *,
    grid_scale_px: float,
    num_octaves: int = 1,
    gain: float = 0.5,
    lacunarity: float = 2.0,
) -> Object:
    object_shape = (1, geometry.height_px, geometry.width_px)
    array = numpy.zeros(object_shape, dtype=complex)
    amplitude = 0.5

    for octave in range(num_octaves):
        noise = generate_simplex_noise_object(rng, geometry, grid_scale_px=grid_scale_px)
        array += amplitude * noise.get_array()
        amplitude *= gain
        grid_scale_px /= lacunarity

    return Object(
        array=array,
        pixel_geometry=geometry.get_pixel_geometry(),
        center=geometry.get_center(),
    )


def generate_dead_leaves_object(
    rng: numpy.random.Generator,
    geometry: ObjectGeometry,
    *,
    leaf_radius_lower_px: float,
    leaf_radius_upper_px: float,
    leaf_radius_power_law_exponent: float,
    leaf_amplitude_lower: float,
    leaf_amplitude_upper: float,
    leaf_phase_lower_tr: float,
    leaf_phase_upper_tr: float,
) -> Object:
    # TODO consider using Poisson disk sampling or CVT for sample positions
    # TODO include pixel aspect ratio in sample position calculation
    _sample_positions = [
        (1.0 / 3, 1.0 / 3),
        (1.0 / 3, 2.0 / 3),
        (2.0 / 3, 1.0 / 3),
        (2.0 / 3, 2.0 / 3),
    ]

    if leaf_radius_lower_px >= leaf_radius_upper_px:
        raise ValueError('leaf_radius_lower_px must be less than leaf_radius_upper_px')

    if leaf_amplitude_lower >= leaf_amplitude_upper:
        raise ValueError('leaf_amplitude_lower must be less than leaf_amplitude_upper')

    if leaf_phase_lower_tr >= leaf_phase_upper_tr:
        raise ValueError('leaf_phase_lower_tr must be less than leaf_phase_upper_tr')

    object_shape = (geometry.height_px, geometry.width_px)
    is_covered = numpy.zeros(object_shape, dtype=bool)
    amplitude: RealArrayType = numpy.zeros(object_shape, dtype=float)
    phase_tr: RealArrayType = numpy.zeros(object_shape, dtype=float)

    position_y_px, position_x_px = numpy.indices(object_shape)
    num_covered_pixels = 0

    beta = 1.0 - leaf_radius_power_law_exponent
    coef = numpy.power(leaf_radius_upper_px / leaf_radius_lower_px, beta) - 1.0

    for leaf in range(100000):  # large value eliminates possiblity of infinite loop
        leaf_radius_px = leaf_radius_lower_px * numpy.power(1.0 + rng.uniform() * coef, 1.0 / beta)
        leaf_position_x_px = rng.uniform(-leaf_radius_px, geometry.width_px + leaf_radius_px)
        leaf_position_y_px = rng.uniform(-leaf_radius_px, geometry.height_px + leaf_radius_px)
        leaf_phase_tr = rng.uniform(
            leaf_phase_lower_tr,
            leaf_phase_upper_tr,
            size=(1, 1),
        )
        leaf_amplitude = numpy.sqrt(
            rng.uniform(
                leaf_amplitude_lower**2,
                leaf_amplitude_upper**2,
                size=(1, 1),
            )
        )

        leaf_counts = numpy.zeros_like(is_covered, dtype=int)
        dx = position_x_px - leaf_position_x_px
        dy = position_y_px - leaf_position_y_px

        # multi-sample anti-aliasing
        for u_y, u_x in _sample_positions:
            sample_distance_px = numpy.hypot(dy + u_y, dx + u_x)
            leaf_counts[sample_distance_px <= leaf_radius_px] += 1

        num_samples = len(_sample_positions)
        leaf_coverage = leaf_counts / num_samples
        amplitude = lerp(amplitude, leaf_amplitude, leaf_coverage)
        phase_tr = lerp(phase_tr, leaf_phase_tr, leaf_coverage)

        is_covered |= leaf_counts == num_samples
        num_covered_pixels = numpy.count_nonzero(is_covered).item()

        covered_pct = 100 * num_covered_pixels / is_covered.size
        logger.info(f'leaves = {leaf}, covered = {covered_pct:.2f}%')

        if num_covered_pixels == is_covered.size:
            break

    array = amplitude * numpy.exp(2j * numpy.pi * phase_tr)

    return Object(
        array=array.astype('complex'),
        pixel_geometry=geometry.get_pixel_geometry(),
        center=geometry.get_center(),
    )


def generate_layers(object_: Object, layer_spacing_m: Sequence[float]) -> Object:
    """Create an object from an existing object with a potentially
    different number of slices.

    If the new object is supposed to be a multislice object with a
    different number of slices than the existing object, the object is
    created as
    `abs(o) ** (1 / nSlices) * exp(i * unwrapPhase(o) / nSlices)`.
    Otherwise, the object is copied as is.
    """
    num_slices = 1 + len(layer_spacing_m)
    array = object_.get_array()

    if num_slices < array.shape[0]:
        array = array[:num_slices]
    elif num_slices > array.shape[0]:
        amplitude = numpy.absolute(array[:1]) ** (1.0 / num_slices)
        amplitude = amplitude.repeat(num_slices, axis=0)
        phase = PhaseUnwrapper().unwrap(array[0])[numpy.newaxis, ...] / num_slices
        phase = phase.repeat(num_slices, axis=0)
        array = amplitude * numpy.exp(1j * phase)

    return Object(
        array=array.astype('complex'),
        pixel_geometry=object_.get_pixel_geometry(),
        center=object_.get_center(),
        layer_spacing_m=layer_spacing_m,
    )


def pad_object(object_: Object, pad_x: int, pad_y: int) -> Object:
    pad_width = [(0, 0), (pad_y, pad_y), (pad_x, pad_x)]

    return Object(
        array=numpy.pad(object_.get_array(), pad_width),
        pixel_geometry=object_.get_pixel_geometry(),
        center=object_.get_center(),
        layer_spacing_m=object_.layer_spacing_m,
    )



from __future__ import annotations

import numpy

from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider
from ptychodus.api.probe_gen import rescale_probe_intensity, generate_super_gaussian_probe

from .builder import ProbeSequenceBuilder
from .settings import ProbeSettings


class SuperGaussianProbeBuilder(ProbeSequenceBuilder):
    def __init__(self, rng: numpy.random.Generator, settings: ProbeSettings) -> None:
        super().__init__(settings, 'super_gaussian')
        self._rng = rng
        self._settings = settings

        self.annular_radius_m = settings.super_gaussian_annular_radius_m.copy()
        self._add_parameter('annular_radius_m', self.annular_radius_m)

        self.fwhm_m = settings.super_gaussian_width_m.copy()
        self._add_parameter('full_width_at_half_maximum_m', self.fwhm_m)

        self.order_parameter = settings.super_gaussian_order_parameter.copy()
        self._add_parameter('order_parameter', self.order_parameter)

    def copy(self) -> SuperGaussianProbeBuilder:
        builder = SuperGaussianProbeBuilder(self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        probe = rescale_probe_intensity(
            generate_super_gaussian_probe(
                geometry_provider.get_probe_geometry(),
                annular_radius_m=self.annular_radius_m.get_value(),
                fwhm_m=self.fwhm_m.get_value(),
                order_parameter=self.order_parameter.get_value(),
            ),
            geometry_provider.probe_photon_count,
        )
        return self._build_probe_modes(self._rng, probe, geometry_provider.num_scan_points)



import logging

import numpy

from ptychodus.api.diffraction import BadPixels
from ptychodus.api.geometry import ImageExtent, PixelGeometry
from ptychodus.api.observer import Observable, Observer
from .settings import DetectorSettings

logger = logging.getLogger(__name__)


class Detector(Observable, Observer):
    def __init__(self, settings: DetectorSettings) -> None:
        super().__init__()
        self._settings = settings
        self._bad_pixels: BadPixels | None = None

        settings.add_observer(self)

    def set_extent(self, extent: ImageExtent) -> None:
        logger.debug(f'Detector {extent=}')
        self._settings.height_px.set_value(extent.height_px)
        self._settings.width_px.set_value(extent.width_px)

    def get_pixel_geometry(self) -> PixelGeometry:
        return PixelGeometry(
            width_m=self._settings.pixel_width_m.get_value(),
            height_m=self._settings.pixel_height_m.get_value(),
        )

    def set_bad_pixels(self, bad_pixels: BadPixels | None) -> None:
        if bad_pixels is not None and bad_pixels.ndim != 2:
            raise ValueError(f'Bad pixels array must be 2D, got {bad_pixels.ndim}D.')

        self._bad_pixels = bad_pixels
        self.notify_observers()

    def get_bad_pixels(self) -> BadPixels:
        if self._bad_pixels is None:
            detector_height_px = self._settings.height_px.get_value()
            detector_width_px = self._settings.width_px.get_value()
            return numpy.full((detector_height_px, detector_width_px), False)

        return self._bad_pixels

    def get_num_bad_pixels(self) -> int:
        return 0 if self._bad_pixels is None else int(numpy.count_nonzero(self._bad_pixels))

    def _update(self, observable: Observable) -> None:
        if observable is self._settings:
            self.notify_observers()



from __future__ import annotations

import numpy

from ptychodus.api.probe_positions import ProbePositionSequence
from ptychodus.api.probe_positions_gen import generate_concentric_probe_positions

from .builder import ProbePositionsBuilder
from .settings import ProbePositionsSettings


class ConcentricProbePositionsBuilder(ProbePositionsBuilder):
    def __init__(self, rng: numpy.random.Generator, settings: ProbePositionsSettings) -> None:
        super().__init__(rng, settings, 'concentric')
        self._settings = settings

        self.radial_step_size_m = settings.radial_step_size_m.copy()
        self._add_parameter('radial_step_size_m', self.radial_step_size_m)

        self.num_shells = settings.num_shells.copy()
        self._add_parameter('num_shells', self.num_shells)

        self.num_points_1st_shell = settings.num_points_in_first_shell.copy()
        self._add_parameter('num_points_1st_shell', self.num_points_1st_shell)

    def copy(self) -> ConcentricProbePositionsBuilder:
        builder = ConcentricProbePositionsBuilder(self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self) -> ProbePositionSequence:
        positions = generate_concentric_probe_positions(
            self.radial_step_size_m.get_value(),
            self.num_shells.get_value(),
            self.num_points_1st_shell.get_value(),
        )
        return self._create_position_sequence(positions)



import logging
import queue

from ptychodus.api.observer import Observable

logger = logging.getLogger(__name__)


class GlobusAuthorizer(Observable):
    def __init__(self) -> None:
        super().__init__()
        self._auth_url_q: queue.Queue[str] = queue.Queue()
        self._auth_code_q: queue.Queue[str] = queue.Queue()
        self._auth_url = ''

    @property
    def needs_authorize_code(self) -> bool:
        return len(self._auth_url) > 0

    def get_authorize_url(self) -> str:
        return self._auth_url

    def set_code_from_authorize_url(self, code: str) -> None:
        logger.info('Received authorization code.')
        self._auth_code_q.put(code)
        self._auth_url = ''
        self.notify_observers()

    def cancel_authorization(self) -> None:
        logger.info('Canceling authorization.')
        self._auth_code_q.put('')
        self._auth_url = ''
        self.notify_observers()

    def run_foreground_tasks(self) -> None:
        try:
            url = self._auth_url_q.get(block=False)
        except queue.Empty:
            pass
        else:
            logger.info(f'Authorize at {url}')
            self._auth_url = url
            self._authorize_code = ''
            self.notify_observers()



from __future__ import annotations
import logging

import numpy

from ptychodus.api.observer import Observable
from ptychodus.api.parametric import ParameterGroup
from ptychodus.api.probe_positions import ProbePositionSequence, ScanGeometry
from ptychodus.api.probe_positions_gen import calculate_scan_geometry

from .builder import FromMemoryProbePositionsBuilder, ProbePositionsBuilder
from .settings import ProbePositionsSettings

logger = logging.getLogger(__name__)


class ProbePositionsRepositoryItem(ParameterGroup):
    def __init__(
        self,
        rng: numpy.random.Generator,
        settings: ProbePositionsSettings,
        builder: ProbePositionsBuilder,
    ) -> None:
        super().__init__()
        self._rng = rng
        self._settings = settings
        self._builder = builder
        self._probe_positions = ProbePositionSequence()
        self._geometry: ScanGeometry | None = None

        self._add_group('builder', builder, observe=True)

        self.expand_bbox = settings.expand_bbox.copy()
        self._add_parameter('expand_bbox', self.expand_bbox)

        self.expand_bbox_xmin_m = settings.expand_bbox_xmin_m.copy()
        self._add_parameter('expand_bbox_xmin_m', self.expand_bbox_xmin_m)

        self.expand_bbox_xmax_m = settings.expand_bbox_xmax_m.copy()
        self._add_parameter('expand_bbox_xmax_m', self.expand_bbox_xmax_m)

        self.expand_bbox_ymin_m = settings.expand_bbox_ymin_m.copy()
        self._add_parameter('expand_bbox_ymin_m', self.expand_bbox_ymin_m)

        self.expand_bbox_ymax_m = settings.expand_bbox_ymax_m.copy()
        self._add_parameter('expand_bbox_ymax_m', self.expand_bbox_ymax_m)

        self._rebuild()

    def assign_item(self, item: ProbePositionsRepositoryItem) -> None:
        self.set_builder(item.get_builder().copy())
        self._rebuild()

    def assign(self, scan: ProbePositionSequence) -> None:
        builder = FromMemoryProbePositionsBuilder(self._rng, self._settings, scan)
        self.set_builder(builder)

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

        self._builder.sync_to_settings()

    def get_probe_positions(self) -> ProbePositionSequence:
        return self._probe_positions

    def get_builder(self) -> ProbePositionsBuilder:
        return self._builder

    def set_builder(self, builder: ProbePositionsBuilder) -> None:
        group = 'builder'
        self._remove_group(group)
        self._builder.remove_observer(self)
        self._builder = builder
        self._builder.add_observer(self)
        self._add_group(group, self._builder, observe=True)
        self._rebuild()

    def get_geometry(self) -> ScanGeometry | None:
        if self.expand_bbox.get_value():
            minimum_x_m = self.expand_bbox_xmin_m.get_value()
            maximum_x_m = self.expand_bbox_xmax_m.get_value()
            minimum_y_m = self.expand_bbox_ymin_m.get_value()
            maximum_y_m = self.expand_bbox_ymax_m.get_value()
            lenth_m = 0.0

            if self._geometry is not None:
                minimum_x_m = min(minimum_x_m, self._geometry.minimum_x_m)
                maximum_x_m = max(maximum_x_m, self._geometry.maximum_x_m)
                minimum_y_m = min(minimum_y_m, self._geometry.minimum_y_m)
                maximum_y_m = max(maximum_y_m, self._geometry.maximum_y_m)
                lenth_m = self._geometry.length_m

            return ScanGeometry(
                minimum_x_m=minimum_x_m,
                maximum_x_m=maximum_x_m,
                minimum_y_m=minimum_y_m,
                maximum_y_m=maximum_y_m,
                length_m=lenth_m,
            )

        return self._geometry

    def _rebuild(self) -> None:
        try:
            probe_positions = self._builder.build()
        except Exception:
            logger.exception('Failed to rebuild scan!')
            return

        self._probe_positions = ProbePositionSequence(probe_positions)
        self._geometry = calculate_scan_geometry(probe_positions)
        self.notify_observers()

    def _update(self, observable: Observable) -> None:
        if observable is self._builder:
            self._rebuild()
        else:
            super()._update(observable)



from __future__ import annotations
from collections.abc import Sequence

import numpy

from ptychodus.api.object import Object, ObjectGeometryProvider
from ptychodus.api.object_gen import generate_fractal_noise_object

from .builder import ObjectBuilder
from .settings import ObjectSettings


class FractalNoiseObjectBuilder(ObjectBuilder):
    def __init__(self, rng: numpy.random.Generator, settings: ObjectSettings) -> None:
        super().__init__(settings, 'fractal_noise')
        self._rng = rng
        self._settings = settings

        self.grid_scale_px = settings.simplex_grid_scale_px.copy()
        self._add_parameter('grid_scale_px', self.grid_scale_px)

        self.num_octaves = settings.fractal_num_octaves.copy()
        self._add_parameter('num_octaves', self.num_octaves)

        self.gain = settings.fractal_gain.copy()
        self._add_parameter('gain', self.gain)

        self.lacunarity = settings.fractal_lacunarity.copy()
        self._add_parameter('lacunarity', self.lacunarity)

    def copy(self) -> FractalNoiseObjectBuilder:
        builder = FractalNoiseObjectBuilder(self._rng, self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(
        self,
        geometry_provider: ObjectGeometryProvider,
        layer_spacing_m: Sequence[float],
    ) -> Object:
        object_ = generate_fractal_noise_object(
            self._rng,
            geometry_provider.get_object_geometry(),
            grid_scale_px=self.grid_scale_px.get_value(),
            num_octaves=self.num_octaves.get_value(),
            gain=self.gain.get_value(),
            lacunarity=self.lacunarity.get_value(),
        )
        return self._create_object(object_, layer_spacing_m)



import numpy

from ptychodus.api.diffraction import BadPixels, DiffractionIndexes, DiffractionPatterns
from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.io import AssembledDiffractionData
from ptychodus.api.propagator import FraunhoferPropagator, PropagatorParameters

from ..diffraction import AssembledDiffractionDataset
from ..product import ProductRepository
from .interpolators import BarycentricArrayInterpolator


class DiffractionSimulator:
    def __init__(self, dataset: AssembledDiffractionDataset, repository: ProductRepository) -> None:
        super().__init__()
        self._dataset = dataset
        self._repository = repository

    def simulate(self, product_index: int) -> None:
        product = self._repository[product_index].get_product()
        # TODO also support multislice
        interpolator = BarycentricArrayInterpolator(product.object_.get_layer(0))
        probe_geometry = product.probes.get_geometry()

        propagator_parameters = PropagatorParameters(
            wavelength_m=product.metadata.probe_wavelength_m,
            width_px=probe_geometry.width_px,
            height_px=probe_geometry.height_px,
            pixel_width_m=probe_geometry.pixel_width_m,
            pixel_height_m=probe_geometry.pixel_height_m,
            propagation_distance_m=product.metadata.detector_distance_m,
        )

        # TODO also support near-field propagation
        propagator = FraunhoferPropagator(propagator_parameters)

        num_positions = len(product.probe_positions)
        indexes: DiffractionIndexes = numpy.array(num_positions, dtype=int)
        patterns: DiffractionPatterns = numpy.zeros(
            (num_positions, probe_geometry.height_px, probe_geometry.width_px),
            dtype=float,
        )
        lambda_z_m2 = (
            propagator_parameters.wavelength_m * propagator_parameters.propagation_distance_m
        )
        pixel_geometry = PixelGeometry(
            width_m=lambda_z_m2 / probe_geometry.width_m,
            height_m=lambda_z_m2 / probe_geometry.height_m,
        )
        bad_pixels: BadPixels = numpy.full(
            (probe_geometry.height_px, probe_geometry.width_px), False
        )

        for index, (probe_position, probe) in enumerate(
            zip(product.probe_positions, product.probes)
        ):
            object_geometry = product.object_.get_geometry()
            object_position = object_geometry.map_coordinates_probe_to_object(probe_position)

            probe_geometry = product.probes.get_geometry()
            object_patch = interpolator.get_patch(
                object_position.coordinate_x_px,
                object_position.coordinate_y_px,
                probe_geometry.width_px,
                probe_geometry.height_px,
            )

            for imode in range(probe.num_incoherent_modes):
                exit_wave = probe.get_incoherent_mode(imode) * object_patch
                wavefield = propagator.propagate(exit_wave)
                patterns[index, :, :] += numpy.square(numpy.abs(wavefield))

            indexes[index] = object_position.index

        data = AssembledDiffractionData(indexes, patterns, pixel_geometry, bad_pixels)
        self._dataset.set_assembled_patterns(data)
